from supervisely._utils import abs_url, is_debug_with_sly_net, is_development
from supervisely.app.widgets import Button, Card, ClassesTable, Container, Text


class ClassesSelector:
    title = "Classes Selector"
    description = (
        "Select classes that will be used for training. "
        "Supported shapes are Bitmap, Polygon, Rectangle."
    )
    lock_message = "Select training and validation splits to unlock"

    def __init__(self, project_id: int, classes: list, app_options: dict = {}):
        self.display_widgets = []

        # GUI Components
        if is_development() or is_debug_with_sly_net():
            qa_stats_link = abs_url(f"projects/{project_id}/stats/datasets")
        else:
            qa_stats_link = f"/projects/{project_id}/stats/datasets"
        qa_stats_text = Text(
            text=f"<i class='zmdi zmdi-chart-donut' style='color: #7f858e'></i> <a href='{qa_stats_link}' target='_blank'> <b> QA & Stats </b></a>"
        )

        self.classes_table = ClassesTable(project_id=project_id)
        if len(classes) > 0:
            self.classes_table.select_classes(classes)
        else:
            self.classes_table.select_all()

        self.validator_text = Text("")
        self.validator_text.hide()
        self.button = Button("Select")
        self.display_widgets.extend(
            [qa_stats_text, self.classes_table, self.validator_text, self.button]
        )
        # -------------------------------- #

        self.container = Container(self.display_widgets)
        self.card = Card(
            title=self.title,
            description=self.description,
            content=self.container,
            lock_message=self.lock_message,
            collapsable=app_options.get("collapsable", False),
        )
        self.card.lock()

    @property
    def widgets_to_disable(self) -> list:
        return [self.classes_table]

    def get_selected_classes(self) -> list:
        return self.classes_table.get_selected_classes()

    def set_classes(self, classes) -> None:
        self.classes_table.select_classes(classes)

    def select_all_classes(self) -> None:
        self.classes_table.select_all()

    def validate_step(self) -> bool:
        self.validator_text.hide()

        if len(self.classes_table.project_meta.obj_classes) == 0:
            self.validator_text.set(text="Project has no classes", status="error")
            self.validator_text.show()
            return False

        selected_classes = self.classes_table.get_selected_classes()
        table_data = self.classes_table._table_data

        empty_classes = [
            row[0]["data"]
            for row in table_data
            if row[0]["data"] in selected_classes and row[2]["data"] == 0 and row[3]["data"] == 0
        ]

        n_classes = len(selected_classes)
        if n_classes == 0:
            self.validator_text.set(text="Please select at least one class", status="error")
        else:
            warning_text = ""
            status = "success"
            if empty_classes:
                intersections = set(selected_classes).intersection(empty_classes)
                if intersections:
                    warning_text = (
                        f". Selected class has no annotations: {', '.join(intersections)}"
                        if len(intersections) == 1
                        else f". Selected classes have no annotations: {', '.join(intersections)}"
                    )
                    status = "warning"

            class_text = "class" if n_classes == 1 else "classes"
            self.validator_text.set(
                text=f"Selected {n_classes} {class_text}{warning_text}", status=status
            )
        self.validator_text.show()
        return n_classes > 0



from __future__ import annotations

import asyncio
import re
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, Callable, Dict, List, Literal, Optional, Union
from uuid import uuid4

from fastapi.responses import HTMLResponse
from pydantic import BaseModel

from supervisely._utils import is_production
from supervisely.api.api import Api
from supervisely.app.widgets import Widget
from supervisely.io.env import task_id
from supervisely.sly_logger import logger


class DebouncedEventHandler:
    def __init__(self, debounce_time: float = 0.1):
        self._event_queue = []
        self._debounce_time = debounce_time
        self._task = None

    async def _process_events(self, func: Callable):
        await asyncio.sleep(self._debounce_time)
        aggregated_events = self._event_queue.copy()
        self._event_queue.clear()
        func(aggregated_events)

    def handle_event(self, event_data, func: Callable):
        self._event_queue.append(event_data)
        if self._task is None or self._task.done():
            self._task = asyncio.create_task(self._process_events(func))


class SelectedIds(BaseModel):
    selected_ids: List[int]
    plot_id: Optional[Union[str, int]] = None


class Bokeh(Widget):
    """
    Bokeh widget for creating interactive plots.

    Note:
        Only Bokeh version 3.1.1 is supported.

    :param plots: List of plots to be displayed.
    :type plots: List[Plot]
    :param width: Width of the chart in pixels.
    :type width: int
    :param height: Height of the chart in pixels.
    :type height: int
    :param tools: List of tools to be displayed on the chart.
    :type tools: List[str]
    :param toolbar_location: Location of the toolbar.
    :type toolbar_location: Literal["above", "below", "left", "right"]
    :param x_axis_visible: If True, x-axis will be visible.
    :type x_axis_visible: bool
    :param y_axis_visible: If True, y-axis will be visible.
    :type y_axis_visible: bool
    :param grid_visible: If True, grid will be visible.
    :type grid_visible: bool
    :param widget_id: Unique widget identifier.
    :type widget_id: str
    :param show_legend: If True, ckickable legend widget will be displayed.
    :type show_legend: bool
    :param legend_location: Location of the clickable legend widget.
    :type legend_location: Literal["left", "top", "right", "bottom"]
    :param legend_click_policy: Click policy of the clickable legend widget.
    :type legend_click_policy: Literal["hide", "mute"]

    :Usage example:
    .. code-block:: python

        from supervisely.app.widgets import Bokeh, IFrame

        plot = Bokeh.Circle(
            x_coordinates=[1, 2, 3, 4, 5],
            y_coordinates=[1, 2, 3, 4, 5],
            radii=10,
            colors="red",
            legend_label="Circle plot",
        )
        bokeh = Bokeh(plots=[plot], width=1000, height=600)

        # To allow the widget to be interacted with, you need to add it to the IFrame widget.
        iframe = IFrame()
        iframe.set(bokeh.html_route_with_timestamp, height="650px", width="100%")
    """

    class Routes:
        VALUE_CHANGED = "value_changed"
        HTML_ROUTE = "bokeh.html"

    class Plot(ABC):
        @abstractmethod
        def add(self, plot) -> None:
            pass

        @abstractmethod
        def register(self, route_path: str) -> None:
            pass

    class Circle(Plot):
        def __init__(
            self,
            x_coordinates: List[Union[int, float]],
            y_coordinates: List[Union[int, float]],
            radii: Optional[Union[Union[int, float], List[Union[int, float]]]] = None,
            colors: Optional[Union[str, List[str]]] = None,
            data: Optional[List[Any]] = None,
            dynamic_selection: bool = False,
            fill_alpha: float = 0.5,
            line_color: Optional[str] = None,
            legend_label: Optional[str] = None,
            plot_id: Optional[Union[str, int]] = None,
        ):
            if not colors:
                colors = Bokeh._generate_colors(x_coordinates, y_coordinates)
            elif isinstance(colors, str):
                colors = [colors] * len(x_coordinates)

            if not radii:
                radii = Bokeh._generate_radii(x_coordinates, y_coordinates)
            elif isinstance(radii, (int, float)):
                radii = [radii] * len(x_coordinates)

            if not len(x_coordinates) == len(y_coordinates) == len(radii) == len(colors):
                raise ValueError(
                    "x_coordinates, y_coordinates, radii, and colors must have the same length"
                )

            if data is not None and len(data) != len(x_coordinates):
                raise ValueError("data must have the same length as x_coordinates")

            if data is None:
                data = list(range(len(x_coordinates)))

            self._x_coordinates = x_coordinates
            self._y_coordinates = y_coordinates
            self._radii = radii
            self._colors = colors
            self._data = data
            self._source = None
            self._dynamic_selection = dynamic_selection
            self._fill_alpha = fill_alpha
            self._line_color = line_color
            self._plot_id = plot_id or uuid4().hex
            self._legend_label = legend_label or str(self._plot_id)

        def add(self, plot) -> None:
            from bokeh.models import (  # pylint: disable=import-error
                ColumnDataSource,
                LassoSelectTool,
            )

            data = dict(
                x=self._x_coordinates,
                y=self._y_coordinates,
                radius=self._radii,
                colors=self._colors,
                ids=self._data,
            )
            self._source = ColumnDataSource(data=data)

            renderer = plot.circle(
                "x",
                "y",
                radius="radius",
                fill_color="colors",
                fill_alpha=self._fill_alpha,
                line_color=self._line_color,
                source=self._source,
                legend_label=self._legend_label,
            )
            if not self._dynamic_selection:
                for tool in plot.tools:
                    if isinstance(tool, (LassoSelectTool)):
                        tool.continuous = False

            return renderer

        def register(self, route_path: str) -> None:
            from bokeh.models import CustomJS  # pylint: disable=import-error

            if not hasattr(self, "_source"):
                raise ValueError("Plot must be added to a Bokeh plot before registering")

            if is_production():
                api = Api()
                task_info = api.task.get_info_by_id(task_id())
                if task_info is not None:
                    route_path = f"/net/{task_info['meta']['sessionToken']}{route_path}"
            callback = CustomJS(
                args=dict(source=self._source),
                code="""
                    var indices = source.selected.indices;
                    var selected_ids = [];
                    for (var i = 0; i < indices.length; i++) {{
                        selected_ids.push(source.data['ids'][indices[i]]);
                    }}
                    var xhr = new XMLHttpRequest();
                    xhr.open("POST", "{route_path}", true);
                    xhr.setRequestHeader("Content-Type", "application/json");
                    xhr.send(JSON.stringify({{selected_ids: selected_ids, plot_id: '{plot_id}'}}));
                """.format(
                    route_path=route_path,
                    plot_id=self._plot_id,
                ),
            )
            self._source.selected.js_on_change("indices", callback)

    def __init__(
        self,
        plots: List[Plot],
        width: int = 1000,
        height: int = 600,
        tools: List[str] = [
            "pan",
            "wheel_zoom",
            "box_zoom",
            "reset",
            "save",
            "poly_select",
            "tap",
            "lasso_select",
        ],
        toolbar_location: Literal["above", "below", "left", "right"] = "above",
        x_axis_visible: bool = False,
        y_axis_visible: bool = False,
        grid_visible: bool = False,
        widget_id: Optional[str] = None,
        show_legend: bool = False,
        legend_location: Literal["left", "top", "right", "bottom"] = "right",
        legend_click_policy: Literal["hide", "mute"] = "hide",
        **kwargs,
    ):
        import bokeh  # pylint: disable=import-error

        # check Bokeh version compatibility (only 3.1.1 is supported)
        if bokeh.__version__ != "3.1.1":
            raise RuntimeError(f"Bokeh version {bokeh.__version__} is not supported. Use 3.1.1")

        self.widget_id = widget_id
        self._plots = plots

        self._width = width
        self._height = height
        self._tools = tools
        self._toolbar_location = toolbar_location
        self._x_axis_visible = x_axis_visible
        self._y_axis_visible = y_axis_visible
        self._grid_visible = grid_visible
        self._show_legend = show_legend
        self._legend_location = legend_location
        self._legend_click_policy = legend_click_policy

        super().__init__(widget_id=widget_id, file_path=__file__)
        self._load_chart()

        server = self._sly_app.get_server()

        @server.get(self.html_route)
        def _html_response() -> None:
            return HTMLResponse(content=self.get_html())

        # TODO: support for offline mode
        # JinjaWidgets().context.pop(self.widget_id, None)  # remove the widget from index.html

    def _load_chart(self, **kwargs):
        from bokeh.models import Legend  # pylint: disable=import-error
        from bokeh.plotting import figure  # pylint: disable=import-error

        self._width = kwargs.get("width", self._width)
        self._height = kwargs.get("height", self._height)
        self._tools = kwargs.get("tools", self._tools)
        self._toolbar_location = kwargs.get("toolbar_location", self._toolbar_location)
        self._show_legend = kwargs.get("show_legend", self._show_legend)
        self._legend_location = kwargs.get("legend_location", self._legend_location)
        self._legend_click_policy = kwargs.get("legend_click_policy", self._legend_click_policy)
        self._x_axis_visible = kwargs.get("x_axis_visible", self._x_axis_visible)
        self._y_axis_visible = kwargs.get("y_axis_visible", self._y_axis_visible)
        self._grid_visible = kwargs.get("grid_visible", self._grid_visible)

        self._plot = figure(
            width=self._width,
            height=self._height,
            tools=self._tools,
            toolbar_location=self._toolbar_location,
        )

        if self._show_legend:
            self._plot.add_layout(
                Legend(click_policy=self._legend_click_policy),
                self._legend_location,
            )

        self._plot.xaxis.visible = self._x_axis_visible
        self._plot.yaxis.visible = self._y_axis_visible
        self._plot.grid.visible = self._grid_visible

        self._renderers = []
        self._process_plots(self._plots)
        self._update_html()

    @property
    def route_path(self) -> str:
        return self.get_route_path(Bokeh.Routes.VALUE_CHANGED)

    @property
    def html_route(self) -> str:
        return self.get_route_path(Bokeh.Routes.HTML_ROUTE)

    @property
    def html_route_with_timestamp(self) -> str:
        return f".{self.html_route}?t={datetime.now().timestamp()}"

    def add_plots(self, plots: List[Plot]) -> None:
        self._plots.extend(plots)
        self._process_plots(plots)
        self._update_html()

    def clear(self) -> None:
        self._plots = []
        self._renderers = []
        self._plot.renderers = []
        self._update_html()

    def remove_plot(self, idx: int) -> None:
        renderer = self._renderers.pop(idx)
        self._plot.renderers.remove(renderer)
        self._update_html()

    def _process_plots(self, plots: List[Plot]) -> None:
        for plot in plots:
            renderer = plot.add(self._plot)
            plot.register(self.route_path)
            self._renderers.append(renderer)

    def _update_html(self) -> None:
        from bokeh.embed import components  # pylint: disable=import-error

        script, self._div = components(self._plot, wrap_script=False)
        self._div_id = self._get_div_id(self._div)
        self._script = self._update_script(script)

    @staticmethod
    def _generate_colors(x_coordinates: List[int], y_coordinates: List[int]) -> List[str]:
        colors = [
            "#%02x%02x%02x" % (int(r), int(g), 150)
            for r, g in zip(
                [50 + 2 * xi for xi in x_coordinates], [30 + 2 * yi for yi in y_coordinates]
            )
        ]
        return colors

    @staticmethod
    def _generate_radii(x_coordinates: List[int], y_coordinates: List[int]) -> List[int]:
        return [1] * len(x_coordinates)

    def _get_div_id(self, div: str) -> str:
        match = re.search(r'id="([^"]+)"', div)
        if match:
            return match.group(1)
        raise ValueError(f"No div id found in {div}")

    def _update_script(self, script: str) -> str:
        # TODO: Reimplement using regex.
        insert_after = "const fn = function() {"
        updated_script = ""
        for line in script.split("\n"):
            if line.strip().startswith(insert_after):
                line = line + f"\n    const el = document.querySelector('#{self._div_id}');"
                line += "\n    if (el === null) {"
                line += "\n      setTimeout(fn, 500);"
                line += "\n      return;"
                line += "\n    }"
            updated_script += line + "\n"
        return updated_script

    def get_json_data(self):
        return {}

    def get_json_state(self):
        return {}

    def value_changed(self, func: Callable) -> Callable:
        server = self._sly_app.get_server()
        self._changes_handled = True
        debounced_handler = DebouncedEventHandler(debounce_time=0.2)  # TODO: check if it's enough

        @server.post(self.route_path)
        async def _click(data: SelectedIds) -> None:
            debounced_handler.handle_event(data, func)

        return _click

    def get_html(self) -> str:
        return f"""<div>
            <script type="text/javascript" src="https://cdn.bokeh.org/bokeh/release/bokeh-3.1.1.min.js"></script>
            <script type="text/javascript"> {self._script} </script>
            {self._div}
        </div>"""

    def update_radii(self, new_radii: Union[List[Union[list, int, float]], int, float]) -> None:
        if isinstance(new_radii, (int, float)):
            new_radii = [new_radii] * len(self._plots)
        elif len(new_radii) != len(self._plots):
            logger.warning(
                f"{len(new_radii)} != {len(self._plots)}: new_radii will be broadcasted to all plots"
            )
            new_radii = [new_radii[0]] * len(self._plots)
        for idx, radii in enumerate(new_radii):
            self.update_radii_by_plot_idx(idx, radii)

    def update_radii_by_plot_idx(self, plot_idx: int, new_radii: List[Union[int, float]]) -> None:
        coords_length = len(self._plots[plot_idx]._x_coordinates)
        if isinstance(new_radii, (int, float)):
            new_radii = [new_radii] * coords_length
        elif len(new_radii) != coords_length:
            logger.warning(
                f"{len(new_radii)} != {coords_length}: new_radii will be broadcasted to all plots"
            )
            new_radii = [new_radii[0]] * coords_length

        self._plots[plot_idx]._radii = new_radii
        self._plots[plot_idx]._source.data["radius"] = new_radii
        self._load_chart()

    def update_chart_size(self, width: Optional[int] = None, height: Optional[int] = None) -> None:
        self._width = width or self._width
        self._height = height or self._height
        self._load_chart()



# coding: utf-8

# docs
from __future__ import annotations

import asyncio
import json
import re
from collections import defaultdict
from typing import (
    AsyncGenerator,
    Callable,
    Dict,
    Generator,
    List,
    NamedTuple,
    Optional,
    Tuple,
    Union,
)

import numpy as np
from requests_toolbelt import MultipartDecoder, MultipartEncoder
from tqdm import tqdm

from supervisely._utils import batched
from supervisely.api.module_api import ApiField, ModuleApi, RemoveableBulkModuleApi
from supervisely.geometry.rectangle import Rectangle
from supervisely.video_annotation.key_id_map import KeyIdMap


class FigureInfo(NamedTuple):
    """
    Represents detailed information about a figure in a scene within the labeling tool.
    It is designed to handle multimodal data.

    Attributes:
        id (int): Unique identifier for the figure.
        class_id (int): Identifier for the class of the figure.
        updated_at (str): Timestamp of the last update.
        created_at (str): Timestamp of creation.
        entity_id (int): Identifier for the entity. Possible entities: image, video, volume, pointcloud, etc.
        object_id (int): Identifier for the object (applicable to videos, volumes, pointclouds).
        project_id (int): Identifier for the project.
        dataset_id (int): Identifier for the dataset.
        frame_index (int): Index of the frame (applicable to videos, volumes (as a slice_index), pointclouds).
        geometry_type (str): Type of geometry.
        geometry (dict): Geometry data.
        geometry_meta (dict): Metadata for the geometry.
        tags (list): List of tags associated with the figure.
        meta (dict): Additional metadata.
        area (str): Area information.
        priority (int): Position of the figure relative to other overlapping or underlying figures.
    """

    id: int
    class_id: int
    updated_at: str
    created_at: str
    entity_id: int
    object_id: int
    project_id: int
    dataset_id: int
    frame_index: int
    geometry_type: str
    geometry: dict
    geometry_meta: dict
    tags: list
    meta: dict
    area: str
    priority: Optional[int] = None

    @property
    def bbox(self) -> Optional[Rectangle]:
        """
        Get Figure's bounding box.

        :return: Rectangle in supervisely format.
        :rtype: :class: `sly.Rectangle`
        """
        if self.geometry_meta is not None:
            return Rectangle(*self.geometry_meta["bbox"], sly_id=self.id)


class FigureApi(RemoveableBulkModuleApi):
    """
    Figure object for :class:`VideoAnnotation<supervisely.video_annotation.video_annotation.VideoAnnotation>`.
    """

    def _remove_batch_api_method_name(self):
        """_remove_batch_api_method_name"""
        return "figures.bulk.remove"

    def _remove_batch_field_name(self):
        """_remove_batch_field_name"""
        return ApiField.FIGURE_IDS

    @staticmethod
    def info_sequence():
        """
        NamedTuple FigureInfo information about Figure.

        :Example:

         .. code-block:: python

            FigureInfo(id=588801373,
                       updated_at='2020-12-22T06:37:13.183Z',
                       created_at='2020-12-22T06:37:13.183Z',
                       entity_id=186648101,
                       object_id=112482,
                       project_id=110366,
                       dataset_id=419886,
                       frame_index=0,
                       geometry_type='bitmap',
                       geometry={'bitmap': {'data': 'eJwdlns8...Cgj4=', 'origin': [335, 205]}})
        """
        return [
            ApiField.ID,
            ApiField.CLASS_ID,
            ApiField.UPDATED_AT,
            ApiField.CREATED_AT,
            ApiField.ENTITY_ID,
            ApiField.OBJECT_ID,
            ApiField.PROJECT_ID,
            ApiField.DATASET_ID,
            ApiField.FRAME_INDEX,
            ApiField.GEOMETRY_TYPE,
            ApiField.GEOMETRY,
            ApiField.GEOMETRY_META,
            ApiField.TAGS,
            ApiField.META,
            ApiField.AREA,
            ApiField.PRIORITY,
        ]

    @staticmethod
    def info_tuple_name():
        """
        Get string name of NamedTuple for class.

        :return: NamedTuple name.
        :rtype: :class:`str`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            tuple_name = api.video.figure.info_tuple_name()
            print(tuple_name) # FigureInfo
        """

        return "FigureInfo"

    def get_info_by_id(self, id: int) -> FigureInfo:
        """
        Get Figure information by ID.

        :param id: Figure ID in Supervisely.
        :type id: int
        :return: Information about Figure. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`NamedTuple`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            figure_id = 588801373

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            figure_info = api.video.figure.get_info_by_id(figure_id)
            print(figure_info)
            # Output: [
            #     588801373,
            #     "2020-12-22T06:37:13.183Z",
            #     "2020-12-22T06:37:13.183Z",
            #     186648101,
            #     112482,
            #     110366,
            #     419886,
            #     0,
            #     "bitmap",
            #     {
            #         "bitmap": {
            #             "data": "eJw...Cgj4=",
            #             "origin": [
            #                 335,
            #                 205
            #             ]
            #         }
            #     }
            # ]
        """
        fields = [
            "id",
            "createdAt",
            "updatedAt",
            "imageId",
            "objectId",
            "classId",
            "projectId",
            "datasetId",
            "geometry",
            "geometryType",
            "geometryMeta",
            "tags",
            "meta",
            "area",
            "priority",
        ]
        return self._get_info_by_id(id, "figures.info", {ApiField.FIELDS: fields})

    def create(
        self,
        entity_id: int,
        object_id: int,
        meta: Dict,
        geometry_json: Dict,
        geometry_type: str,
        track_id: int = None,
    ) -> int:
        """"""
        input_figure = {
            ApiField.META: meta,
            ApiField.OBJECT_ID: object_id,
            ApiField.GEOMETRY_TYPE: geometry_type,
            ApiField.GEOMETRY: geometry_json,
        }

        if track_id is not None:
            input_figure[ApiField.TRACK_ID] = track_id

        body = {ApiField.ENTITY_ID: entity_id, ApiField.FIGURES: [input_figure]}

        response = self._api.post("figures.bulk.add", body)
        return response.json()[0][ApiField.ID]

    def get_by_ids(self, dataset_id: int, ids: List[int]) -> List[FigureInfo]:
        """
        Get Figures information by IDs from given dataset ID.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param ids: List of Figures IDs.
        :type ids: List[int]
        :return: List of information about Figures. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[NamedTuple]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            dataset_id = 466642
            figures_ids = [642155547, 642155548, 642155549]
            figures_infos = api.video.figure.get_by_ids(dataset_id, figures_ids)
            print(figures_infos)
            # Output: [
            #     [
            #         642155547,
            #         "2021-03-23T13:25:34.705Z",
            #         "2021-03-23T13:25:34.705Z",
            #         198703211,
            #         152118,
            #         124976,
            #         466642,
            #         0,
            #         "rectangle",
            #         {
            #             "points": {
            #                 "exterior": [
            #                     [
            #                         2240,
            #                         1041
            #                     ],
            #                     [
            #                         2463,
            #                         1187
            #                     ]
            #                 ],
            #                 "interior": []
            #             }
            #         }
            #     ],
            #     [
            #         642155548,
            #         "2021-03-23T13:25:34.705Z",
            #         "2021-03-23T13:25:34.705Z",
            #         198703211,
            #         152118,
            #         124976,
            #         466642,
            #         1,
            #         "rectangle",
            #         {
            #             "points": {
            #                 "exterior": [
            #                     [
            #                         2248,
            #                         1048
            #                     ],
            #                     [
            #                         2455,
            #                         1176
            #                     ]
            #                 ],
            #                 "interior": []
            #             }
            #         }
            #     ],
            #     [
            #         642155549,
            #         "2021-03-23T13:25:34.705Z",
            #         "2021-03-23T13:25:34.705Z",
            #         198703211,
            #         152118,
            #         124976,
            #         466642,
            #         2,
            #         "rectangle",
            #         {
            #             "points": {
            #                 "exterior": [
            #                     [
            #                         2237,
            #                         1046
            #                     ],
            #                     [
            #                         2464,
            #                         1179
            #                     ]
            #                 ],
            #                 "interior": []
            #             }
            #         }
            #     ]
            # ]
        """
        filters = [{"field": "id", "operator": "in", "value": ids}]
        fields = [
            "id",
            "createdAt",
            "updatedAt",
            "imageId",
            "objectId",
            "classId",
            "projectId",
            "datasetId",
            "geometry",
            "geometryType",
            "geometryMeta",
            "tags",
            "meta",
            "area",
            "priority",
        ]
        figures_infos = self.get_list_all_pages(
            "figures.list",
            {
                ApiField.DATASET_ID: dataset_id,
                ApiField.FILTER: filters,
                ApiField.FIELDS: fields,
            },
        )

        if len(ids) != len(figures_infos):
            ids_downloaded = [info.id for info in figures_infos]
            raise RuntimeError(
                "Ids don't exist on server: {}".format(set(ids_downloaded) - set(ids))
            )

        id_to_item = {info.id: info for info in figures_infos}

        figures = []
        for input_id in ids:
            figures.append(id_to_item[input_id])

        return figures

    def _append_bulk(
        self,
        entity_id,
        figures_json,
        figures_keys,
        key_id_map: KeyIdMap,
        field_name=ApiField.ENTITY_ID,
    ):
        """"""
        if len(figures_json) == 0:
            return
        for batch_keys, batch_jsons in zip(
            batched(figures_keys, batch_size=100), batched(figures_json, batch_size=100)
        ):
            resp = self._api.post(
                "figures.bulk.add",
                {field_name: entity_id, ApiField.FIGURES: batch_jsons},
            )
            for key, resp_obj in zip(batch_keys, resp.json()):
                figure_id = resp_obj[ApiField.ID]
                key_id_map.add_figure(key, figure_id)

    def create_bulk(
        self,
        figures_json: List[dict],
        entity_id: int = None,
        dataset_id: int = None,
        batch_size=200,
    ) -> List[int]:
        """
        Create figures in Supervisely in bulk.
        To optimize creation of a large number of figures use dataset ID instead of entity ID.
        In this case figure jsons list can contain figures from different entities for the same dataset.
        Every figure json must contain corresponding entity ID.

        *NOTE*: Geometries for AlphaMask must be uploaded separately via `upload_geometries_batch` method.

        :param figures_json: List of figures in Supervisely JSON format.
        :type figures_json: List[dict]
        :param entity_id: Entity ID.
        :type entity_id: int
        :param dataset_id: Dataset ID.
        :type dataset_id: int
        :return: List of figure IDs.
        """
        figure_ids = []
        if len(figures_json) == 0:
            return figure_ids

        if entity_id is None and dataset_id is None:
            raise ValueError("Either entity_id or dataset_id must be provided")
        if dataset_id is not None:
            body = {ApiField.DATASET_ID: dataset_id}
        else:
            body = {ApiField.ENTITY_ID: entity_id}

        for batch_jsons in batched(figures_json, batch_size):
            body[ApiField.FIGURES] = batch_jsons
            response = self._api.post("figures.bulk.add", body)
            for resp_obj in response.json():
                figure_ids.append(resp_obj[ApiField.ID])
        return figure_ids

    def download(
        self, dataset_id: int, image_ids: List[int] = None, skip_geometry: bool = False
    ) -> Dict[int, List[FigureInfo]]:
        """
        Method returns a dictionary with pairs of image ID and list of FigureInfo for the given dataset ID. Can be filtered by image IDs.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param image_ids: Specify the list of image IDs within the given dataset ID. If image_ids is None, the method returns all possible pairs of images with figures. Note: Consider using `sly.batched()` to ensure that no figures are lost in the response.
        :type image_ids: List[int], optional
        :param skip_geometry: Skip the download of figure geometry. May be useful for a significant api request speed increase in the large datasets.
        :type skip_geometry: bool

        :return: A dictionary where keys are image IDs and values are lists of figures.
        :rtype: :class: `Dict[int, List[FigureInfo]]`
        """
        fields = [
            "id",
            "createdAt",
            "updatedAt",
            "imageId",
            "objectId",
            "classId",
            "projectId",
            "datasetId",
            "geometry",
            "geometryType",
            "geometryMeta",
            "tags",
            "meta",
            "area",
            "priority",
        ]
        if skip_geometry is True:
            fields = [x for x in fields if x != "geometry"]

        if image_ids is None:
            filters = []
        else:
            filters = [
                {
                    ApiField.FIELD: ApiField.ENTITY_ID,
                    "operator": "in",
                    "value": image_ids,
                }
            ]
        data = {
            ApiField.DATASET_ID: dataset_id,
            ApiField.FIELDS: fields,
            ApiField.FILTER: filters,
        }
        resp = self._api.post("figures.list", data)
        infos = resp.json()
        images_figures = defaultdict(list)

        total_pages = infos["pagesCount"]
        for page in range(1, total_pages + 1):
            if page > 1:
                data.update({ApiField.PAGE: page})
                resp = self._api.post("figures.list", data)
                infos = resp.json()
            for info in infos["entities"]:
                figure_info = self._convert_json_info(info, True)
                images_figures[figure_info.entity_id].append(figure_info)

        return dict(images_figures)

    def _convert_json_info(self, info: dict, skip_missing=False):
        res = super()._convert_json_info(info, skip_missing=True)
        return FigureInfo(**res._asdict())

    def _download_geometries_generator(
        self, ids: List[int]
    ) -> Generator[Tuple[int, MultipartDecoder.Part], None, None]:
        """
        Private method. Download figures geometries with given IDs from storage.
        """

        for batch_ids in batched(ids):
            response = self._api.post("figures.bulk.download.geometry", {ApiField.IDS: batch_ids})
            decoder = MultipartDecoder.from_response(response)
            for part in decoder.parts:
                content_utf8 = part.headers[b"Content-Disposition"].decode("utf-8")
                # Find name="1245" preceded by a whitespace, semicolon or beginning of line.
                # The regex has 2 capture group: one for the prefix and one for the actual name value.
                figure_id = int(re.findall(r'(^|[\s;])name="(\d*)"', content_utf8)[0][1])
                yield figure_id, part

    def download_geometry(self, figure_id: int) -> dict:
        """
        Download figure geometry with given ID from storage.

        :param figure_id: Figure ID in Supervisely.
        :type figure_id: int
        :return: Figure geometry in Supervisely JSON format.
        :rtype: dict
        """
        return self.download_geometries_batch([figure_id])

    def download_geometries_batch(
        self,
        ids: List[int],
        progress_cb: Optional[Union[tqdm, Callable]] = None,
    ) -> List[dict]:
        """
        Download figure geometries with given IDs from storage.

        :param ids: List of figure IDs in Supervisely.
        :type ids: List[int]
        :param progress_cb: Progress bar to show the download progress. Shows the number of bytes downloaded.
        :type progress_cb: Union[tqdm, Callable], optional
        :return: List of figure geometries in Supervisely JSON format.
        :rtype: List[dict]
        """
        geometries = {}
        for idx, part in self._download_geometries_generator(ids):
            if progress_cb is not None:
                progress_cb(len(part.content))
            geometry_json = json.loads(part.content)
            geometries[idx] = geometry_json

        if len(geometries) != len(ids):
            raise RuntimeError("Not all geometries were downloaded")
        ordered_results = [geometries[i] for i in ids]
        return ordered_results

    def upload_geometry(self, figure_id: int, geometry: dict):
        """
        Upload figure geometry with given figure ID to storage.

        :param figure_id: Figure ID in Supervisely.
        :type figure_id: int
        :param geometry: Figure geometry in Supervisely JSON format.
        :type geometry: dict
        :return: None
        :rtype: None
        """
        self.upload_geometries_batch([figure_id], [geometry])

    def upload_geometries_batch(self, figure_ids: List[int], geometries: List[dict]):
        """
        Upload figure geometries with given figure IDs to storage.

        :param figure_ids: List of figure IDs in Supervisely.
        :type figure_ids: List[int]
        :param geometries: List of figure geometries in Supervisely JSON format.
        :type geometries: List[dict]
        :return: None
        :rtype: None
        """
        geometries = [json.dumps(geometry).encode("utf-8") for geometry in geometries]

        for batch_ids, batch_geometries in zip(
            batched(figure_ids, batch_size=100), batched(geometries, batch_size=100)
        ):
            fields = []
            for figure_id, geometry in zip(batch_ids, batch_geometries):
                fields.append((ApiField.FIGURE_ID, str(figure_id)))
                fields.append(
                    (
                        ApiField.GEOMETRY,
                        (str(figure_id), geometry, "application/octet-stream"),
                    )
                )
            encoder = MultipartEncoder(fields=fields)
            self._api.post("figures.bulk.upload.geometry", encoder)

    async def _download_geometries_generator_async(
        self, ids: List[int], semaphore: Optional[asyncio.Semaphore] = None
    ) -> AsyncGenerator[Tuple[int, bytes], None, None]:
        """
        Private method. Download figures geometries with given IDs from storage asynchronously.

        :param ids: List of figure IDs in Supervisely.
        :type ids: List[int]
        :return: Async generator with pairs of figure ID and figure geometry.
        :rtype: AsyncGenerator[Tuple[int, bytes], None, None]
        """
        if semaphore is None:
            semaphore = self._api.get_default_semaphore()

        for batch_ids in batched(ids):
            async with semaphore:
                response = await self._api.post_async(
                    "figures.bulk.download.geometry", {ApiField.IDS: batch_ids}
                )
            decoder = MultipartDecoder.from_response(response)
            for part in decoder.parts:
                content_utf8 = part.headers[b"Content-Disposition"].decode("utf-8")
                # Find name="1245" preceded by a whitespace, semicolon or beginning of line.
                # The regex has 2 capture group: one for the prefix and one for the actual name value.
                figure_id = int(re.findall(r'(^|[\s;])name="(\d*)"', content_utf8)[0][1])
                yield figure_id, part.content

    async def download_geometries_batch_async(
        self,
        ids: List[int],
        progress_cb: Optional[Union[tqdm, Callable]] = None,
        semaphore: Optional[asyncio.Semaphore] = None,
    ) -> List[dict]:
        """
        Download figure geometries with given IDs from storage asynchronously.

        :param ids: List of figure IDs in Supervisely.
        :type ids: List[int]
        :param progress_cb: Progress bar to show the download progress. Shows the number of bytes downloaded.
        :type progress_cb: Union[tqdm, Callable], optional
        :param semaphore: Semaphore to limit the number of concurrent downloads.
        :type semaphore: Optional[asyncio.Semaphore], optional
        :return: List of figure geometries in Supervisely JSON format.
        :rtype: List[dict]

        :Usage example:

            .. code-block:: python

                import asyncio
                import supervisely as sly

                os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
                os.environ['API_TOKEN'] = 'Your Supervisely API Token'
                api = sly.Api.from_env()

                figure_ids = [642155547, 642155548, 642155549]
                loop = sly.utils.get_or_create_event_loop()
                geometries = loop.run_until_complete(
                    api.figure.download_geometries_batch_async(
                        figure_ids,
                        progress_cb=tqdm(total=len(figure_ids), desc="Downloading geometries"),
                        semaphore=asyncio.Semaphore(15),
                    )
                )
        """
        geometries = {}
        async for idx, part in self._download_geometries_generator_async(ids, semaphore):
            if progress_cb is not None:
                progress_cb(len(part))
            geometry_json = json.loads(part)
            geometries[idx] = geometry_json

        if len(geometries) != len(ids):
            raise RuntimeError("Not all geometries were downloaded")
        ordered_results = [geometries[i] for i in ids]
        return ordered_results



from supervisely.nn.benchmark.object_detection.vis_metrics.confidence_distribution import (
    ConfidenceDistribution,
)
from supervisely.nn.benchmark.object_detection.vis_metrics.confidence_score import (
    ConfidenceScore,
)
from supervisely.nn.benchmark.object_detection.vis_metrics.confusion_matrix import (
    ConfusionMatrix,
)
from supervisely.nn.benchmark.object_detection.vis_metrics.explore_predictions import (
    ExplorePredictions,
)
from supervisely.nn.benchmark.object_detection.vis_metrics.f1_score_at_different_iou import (
    F1ScoreAtDifferentIOU,
)
from supervisely.nn.benchmark.object_detection.vis_metrics.frequently_confused import (
    FrequentlyConfused,
)
from supervisely.nn.benchmark.object_detection.vis_metrics.iou_distribution import (
    IOUDistribution,
)
from supervisely.nn.benchmark.object_detection.vis_metrics.key_metrics import KeyMetrics
from supervisely.nn.benchmark.object_detection.vis_metrics.model_predictions import (
    ModelPredictions,
)
from supervisely.nn.benchmark.object_detection.vis_metrics.outcome_counts import (
    OutcomeCounts,
)
from supervisely.nn.benchmark.object_detection.vis_metrics.outcome_counts_per_class import (
    PerClassOutcomeCounts,
)
from supervisely.nn.benchmark.object_detection.vis_metrics.overview import Overview
from supervisely.nn.benchmark.object_detection.vis_metrics.pr_curve import PRCurve
from supervisely.nn.benchmark.object_detection.vis_metrics.pr_curve_by_class import (
    PRCurveByClass,
)
from supervisely.nn.benchmark.object_detection.vis_metrics.precision import Precision
from supervisely.nn.benchmark.object_detection.vis_metrics.precision_avg_per_class import (
    PerClassAvgPrecision,
)
from supervisely.nn.benchmark.object_detection.vis_metrics.recall import Recall
from supervisely.nn.benchmark.object_detection.vis_metrics.recall_vs_precision import (
    RecallVsPrecision,
)
from supervisely.nn.benchmark.object_detection.vis_metrics.reliability_diagram import (
    ReliabilityDiagram,
)
from supervisely.nn.benchmark.object_detection.vis_metrics.speedtest import Speedtest



import asyncio
import os
import time

import supervisely as sly

LOG_LEVEL = "INFO"
# LOG_LEVEL = "DEBUG"
api = sly.Api.from_env()

api.logger.setLevel(LOG_LEVEL)

TEAM_ID = 567
user_path = os.path.expanduser("~")
save_path = f"{user_path}/Work/test_files_download/"
sly.logger.info(f"Save path: {save_path}")
remote_path = "/videos/"
sly.fs.ensure_base_path(save_path)
sly.fs.clean_dir(save_path)
files = (
    ("12391768_3840_2160_30fps.mp4", f"{remote_path}12391768_3840_2160_30fps.mp4"),
    ("MOV_h264.mov", f"{remote_path}MOV_h264.mov"),
    ("MP4_HEVC.mp4", f"{remote_path}MP4_HEVC.mp4"),
)


async def download_files():
    sema = asyncio.Semaphore(10)
    tasks = []
    for name, path in files:
        progress = sly.tqdm_sly(total=None, desc=f"Downloading {name}", unit="B", unit_scale=True)
        task = api.file.download_async(TEAM_ID, path, save_path + name, sema, progress_cb=progress)
        tasks.append(task)
    await asyncio.gather(*tasks)


def main_db():
    loop = sly.utils.get_or_create_event_loop()

    remote_paths = [path for _, path in files]
    names = [name for name, _ in files]
    sizeb_list = [api.file.get_info_by_path(TEAM_ID, path).sizeb for _, path in files]
    sizeb = sum(sizeb_list)
    save_paths = [save_path + name for name in names]
    progress = sly.tqdm_sly(total=sizeb, desc="Downloading files", unit="B", unit_scale=True)
    download_coro = api.file.download_bulk_async(
        TEAM_ID, remote_paths, save_paths, progress_cb=progress
    )

    if loop.is_running():
        future = asyncio.run_coroutine_threadsafe(download_coro, loop=loop)
        future.result()
    else:
        loop.run_until_complete(download_coro)


def maind_df():
    loop = asyncio.get_event_loop()
    loop.run_until_complete(download_files())


def main_ip():
    start = time.time()
    loop = asyncio.get_event_loop()
    # os.environ["CONTEXT_SLYFILE"] = "/test/test_input.tar"
    os.environ["FOLDER"] = "/videos/"
    os.environ["TEAM_ID"] = str(TEAM_ID)
    loop.run_until_complete(api.file.download_input_async(save_path, show_progress=True))
    end = time.time()
    print(f"Time taken for download input async: {end-start}")


def main_dd():
    start = time.time()
    loop = asyncio.get_event_loop()
    loop.run_until_complete(
        api.file.download_directory_async(
            TEAM_ID, remote_path, save_path, semaphore=asyncio.Semaphore(100)
        )
    )
    end = time.time()
    print(f"Time taken for download dir async: {end-start}")


def compare_dir_download():
    main_dd()

    sly.fs.clean_dir(save_path)

    start = time.time()
    api.file.download_directory(TEAM_ID, remote_path, save_path)
    end = time.time()
    print(f"Time taken for download dir: {end-start}")


if __name__ == "__main__":
    try:
        # maind_df()  # to download and save files
        # main_dd()  # to download and save files as folder
        main_db()  # to download and save files in bulk
        # main_ip()  # to download input file
        # compare_dir_download()  # to compare download time between async and sync
    except KeyboardInterrupt:
        sly.logger.info("Stopped by user")



import datetime
import json
import os
import random
import re
import shutil
import string
import subprocess
import tarfile
from pathlib import Path

import git
import requests
from giturlparse import parse
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
from tqdm import tqdm


def slug_is_valid(slug):
    splitted = slug.split("/")
    return len(splitted) == 2 and len(splitted[0]) > 0 and len(splitted[1]) > 0


def get_module_root(path: str):
    if path is None:
        return Path(os.getcwd())
    return Path(path).absolute().resolve()


def get_module_path(module_root, sub_app):
    if sub_app is None:
        return module_root
    return module_root.joinpath(sub_app)


def get_remote_url(remote: git.Remote):
    p = parse(remote.url)
    url = p.url2https.replace("https://", "").replace(".git", "").lower()
    return url


def get_semver(string):
    return re.match("\d+\.\d+\.\d+", string)


def find_tag_in_repo(tag_name, repo: git.Repo):
    for tag in repo.tags:
        if tag.name == tag_name:
            return tag
    return None


def push_tag(tag_name, repo: git.Repo):
    remote_name = repo.active_branch.tracking_branch().remote_name
    command = f"git push --porcelain -- {remote_name} {tag_name}"
    try:
        subprocess.check_call(command, shell=True, cwd=repo.working_dir)
    except subprocess.CalledProcessError:
        raise


def delete_tag(tag_name, repo: git.Repo):
    tag = find_tag_in_repo(tag_name, repo)
    repo.delete_tag(tag)
    remote_name = remote_name = repo.active_branch.tracking_branch().remote_name
    command = f"git push --porcelain -- {remote_name} :refs/tags/{tag_name}"
    try:
        subprocess.check_call(command, shell=True, cwd=repo.working_dir)
    except subprocess.CalledProcessError:
        raise


def get_appKey(repo, sub_app_path, repo_url):
    import hashlib

    p = parse(repo_url)
    repo_url = p.url2https.replace("https://", "").replace(".git", "").lower()

    first_commit = next(repo.iter_commits("HEAD", reverse=True))
    key_string = repo_url + "_" + first_commit.hexsha
    appKey = hashlib.md5(key_string.encode("utf-8")).hexdigest()
    if sub_app_path is not None:
        appKey += "_" + hashlib.md5(sub_app_path.encode("utf-8")).hexdigest()
    appKey += "_" + hashlib.md5(first_commit.hexsha[:7].encode("utf-8")).hexdigest()

    return appKey


def get_instance_version(token, server):
    headers = {
        "x-api-key": token,
        "Content-Type": "application/json",
    }
    r = requests.post(f'{server.rstrip("/")}/public/api/v3/instance.version', headers=headers)
    if r.status_code == 403:
        raise PermissionError()
    if r.status_code == 404:
        raise NotImplementedError()
    if r.status_code != 200:
        raise ConnectionError()
    return r.json()


def get_app_from_instance(appKey: str, token, server):
    headers = {
        "x-api-key": token,
        "Content-Type": "application/json",
    }
    data = json.dumps(
        {
            "appKey": appKey,
        }
    )
    r = requests.post(
        f'{server.rstrip("/")}/public/api/v3/ecosystem.info', headers=headers, data=data
    )
    if r.status_code == 403:
        raise PermissionError()
    if r.status_code == 404:
        return None
    if r.status_code != 200:
        raise ConnectionError()
    return r.json()


def upload_archive(
    archive_path,
    server_address,
    api_token,
    appKey,
    release,
    config,
    readme,
    modal_template,
    slug,
    user_id,
    subapp_path,
    share_app,
):
    f = open(archive_path, "rb")
    archive_name = os.path.basename(archive_path)
    fields = {
        "appKey": appKey,
        "subAppPath": subapp_path,
        "release": json.dumps(release),
        "config": json.dumps(config),
        "readme": readme,
        "modalTemplate": modal_template,
        "archive": (
            archive_name,
            f,
            "application/gzip" if archive_name.endswith(".tar.gz") else "application/x-tar",
        ),
    }
    if slug:
        fields["slug"] = slug
    if user_id:
        fields["userId"] = str(user_id)
    if share_app:
        fields["isShared"] = "true"
    e = MultipartEncoder(fields=fields)
    encoder_len = e.len
    with tqdm(
        total=encoder_len,
        unit="B",
        unit_scale=True,
        unit_divisor=1024,
    ) as bar:
        m = MultipartEncoderMonitor(e, lambda monitor: bar.update(monitor.bytes_read - bar.n))
        response = requests.post(
            f"{server_address.rstrip('/')}/public/api/v3/ecosystem.release",
            data=m,
            headers={"Content-Type": m.content_type, "x-api-key": api_token},
        )
    f.close()
    return response


def archive_application(repo: git.Repo, config, slug):
    archive_folder = "".join(random.choice(string.ascii_letters) for _ in range(5))
    os.mkdir(archive_folder)
    file_paths = [
        Path(line.decode("utf-8")).absolute()
        for line in subprocess.check_output(
            "git ls-files --recurse-submodules", shell=True
        ).splitlines()
    ]
    if slug is None:
        app_folder_name = config["name"].lower()
    else:
        app_folder_name = slug.split("/")[1].lower()
    app_folder_name = re.sub("[ \/]", "-", app_folder_name)
    app_folder_name = re.sub("[\"'`,\[\]\(\)]", "", app_folder_name)
    working_dir_path = Path(repo.working_dir).absolute()
    if config.get("type", "app") == "client_side_app":
        archive_path = archive_folder + "/archive.tar"
        write_mode = "w"
    else:
        archive_path = archive_folder + "/archive.tar.gz"
        write_mode = "w:gz"
    with tarfile.open(archive_path, write_mode) as tar:
        for path in file_paths:
            if path.is_file():
                tar.add(
                    path.absolute(),
                    Path(app_folder_name).joinpath(path.relative_to(working_dir_path)),
                )
    return archive_path


def get_user(server_address, api_token):
    headers = {
        "x-api-key": api_token,
        "Content-Type": "application/json",
    }
    r = requests.post(f'{server_address.rstrip("/")}/public/api/v3/users.me', headers=headers)
    if r.status_code == 403:
        raise PermissionError()
    if r.status_code == 404 or r.status_code == 400:
        return None
    if r.status_code != 200:
        raise ConnectionError()
    return r.json()


def delete_directory(path):
    shutil.rmtree(path)


def get_created_at(repo: git.Repo, tag_name):
    if tag_name is None:
        return None
    for tag in repo.tags:
        if tag.name == tag_name:
            if tag.tag is None:
                timestamp = tag.commit.committed_date
            else:
                timestamp = tag.tag.tagged_date
            return datetime.datetime.utcfromtimestamp(timestamp).isoformat()
    return None


def release(
    server_address,
    api_token,
    appKey,
    repo: git.Repo,
    config,
    readme,
    release_name,
    release_version,
    modal_template="",
    slug=None,
    user_id=None,
    subapp_path="",
    created_at=None,
    share_app=False,
):
    if created_at is None:
        created_at = get_created_at(repo, release_version)
    archive_path = archive_application(repo, config, slug)
    release = {
        "name": release_name,
        "version": release_version,
    }
    if created_at is not None:
        release["createdAt"] = created_at
    try:
        response = upload_archive(
            archive_path,
            server_address,
            api_token,
            appKey,
            release,
            config,
            readme,
            modal_template,
            slug,
            user_id,
            subapp_path,
            share_app,
        )
    finally:
        delete_directory(os.path.dirname(archive_path))
    return response



from typing import Any, Dict, List, Union

from supervisely.annotation.label import Label
from supervisely.annotation.tag import Tag
from supervisely.geometry.bitmap import Bitmap, Rectangle
from supervisely.nn.inference.inference import Inference
from supervisely.nn.prediction_dto import PredictionBBox, PredictionMask
from supervisely.sly_logger import logger


class InstanceSegmentation(Inference):
    def get_info(self) -> dict:
        info = super().get_info()
        info["task type"] = "instance segmentation"
        # recommended parameters:
        # info["model_name"] = ""
        # info["checkpoint_name"] = ""
        # info["pretrained_on_dataset"] = ""
        # info["device"] = ""
        return info

    def _get_obj_class_shape(self):
        return Bitmap

    def _create_label(self, dto: Union[PredictionMask, PredictionBBox]):
        obj_class = self.model_meta.get_obj_class(dto.class_name)
        if obj_class is None:
            raise KeyError(
                f"Class {dto.class_name} not found in model classes {self.get_classes()}"
            )
        if isinstance(dto, PredictionMask):
            if not dto.mask.any():  # skip empty masks
                logger.debug(f"Mask of class {dto.class_name} is empty and will be skipped")
                return None
            geometry = Bitmap(dto.mask, extra_validation=False)
        elif isinstance(dto, PredictionBBox):
            geometry = Rectangle(*dto.bbox_tlbr)
        else:
            raise TypeError(f"Unknown prediction type {type(dto)}")
        tags = []
        if dto.score is not None:
            tags.append(Tag(self._get_confidence_tag_meta(), dto.score))
        label = Label(geometry, obj_class, tags)
        return label

    def predict(
        self, image_path: str, settings: Dict[str, Any]
    ) -> List[Union[PredictionMask, PredictionBBox]]:
        raise NotImplementedError("Have to be implemented in child class")

    def predict_raw(
        self, image_path: str, settings: Dict[str, Any]
    ) -> List[Union[PredictionMask, PredictionBBox]]:
        raise NotImplementedError(
            "Have to be implemented in child class If sliding_window_mode is 'advanced'."
        )



from __future__ import annotations

from typing import Dict

from supervisely.nn.benchmark.object_detection.base_vis_metric import DetectionVisMetric
from supervisely.nn.benchmark.visualization.widgets import (
    ChartWidget,
    MarkdownWidget,
    TableWidget,
)


class KeyMetrics(DetectionVisMetric):
    MARKDOWN = "key_metrics"
    CHART = "key_metrics"
    TABLE = "key_metrics"

    @property
    def md(self) -> MarkdownWidget:
        text = self.vis_texts.markdown_key_metrics.format(
            self.vis_texts.definitions.iou_threshold,
            self.vis_texts.definitions.average_precision,
            self.vis_texts.definitions.confidence_score,
        )
        return MarkdownWidget(self.MARKDOWN, "Key Metrics", text)

    @property
    def table(self) -> TableWidget:
        columns = ["metrics", "values"]
        content = []
        for metric, value in self.eval_result.mp.metric_table().items():
            row = [metric, round(value, 2)]
            dct = {
                "row": row,
                "id": metric,
                "items": row,
            }
            content.append(dct)

        columns_options = [
            {"disableSort": True},  # , "ustomCell": True},
            {"disableSort": True},
        ]

        data = {
            "columns": columns,
            "columnsOptions": columns_options,
            "content": content,
        }
        table = TableWidget(
            name=self.TABLE,
            data=data,
            fix_columns=1,
            width="60%",
            show_header_controls=False,
            main_column=columns[0],
        )
        return table

    @property
    def chart(self) -> ChartWidget:
        return ChartWidget(self.CHART, self._get_figure())

    def _get_figure(self):  # -> go.Figure:
        import plotly.graph_objects as go  # pylint: disable=import-error

        # Overall Metrics
        base_metrics = self.eval_result.mp.base_metrics()
        r = list(base_metrics.values())
        theta = [self.eval_result.mp.metric_names[k] for k in base_metrics.keys()]
        fig = go.Figure()
        fig.add_trace(
            go.Scatterpolar(
                r=r + [r[0]],
                theta=theta + [theta[0]],
                # fill="toself",
                name="Overall Metrics",
                hovertemplate="%{theta}: %{r:.2f}<extra></extra>",
            )
        )
        fig.update_layout(
            polar=dict(
                radialaxis=dict(
                    range=[0.0, 1.0],
                    ticks="outside",
                ),
                angularaxis=dict(rotation=90, direction="clockwise"),
            ),
            dragmode=False,
            margin=dict(l=25, r=25, t=25, b=25),
        )
        fig.update_layout(
            modebar=dict(
                remove=[
                    "zoom2d",
                    "pan2d",
                    "select2d",
                    "lasso2d",
                    "zoomIn2d",
                    "zoomOut2d",
                    "autoScale2d",
                    "resetScale2d",
                ]
            )
        )
        return fig

    def get_click_data(self) -> Dict:
        if not self.clickable:
            return
        res = {}

        res["layoutTemplate"] = [None, None, None]
        res["clickData"] = {}
        for outcome, matches_data in self.eval_result.click_data.outcome_counts.items():
            res["clickData"][outcome] = {}
            res["clickData"][outcome]["imagesIds"] = []

            img_ids = set()
            for match_data in matches_data:
                pairs_data = self.eval_result.matched_pair_data[match_data["gt_img_id"]]
                if outcome == "FN":
                    img_ids.add(pairs_data.diff_image_info.id)
                else:
                    img_ids.add(pairs_data.pred_image_info.id)

            res["clickData"][outcome][
                "title"
            ] = f"{outcome}: {len(matches_data)} object{'s' if len(matches_data) > 1 else ''}"
            res["clickData"][outcome]["imagesIds"] = list(img_ids)
            res["clickData"][outcome]["filters"] = [
                {"type": "tag", "tagId": "confidence", "value": [0, 1]},
                {"type": "tag", "tagId": "outcome", "value": outcome},
            ]

        return res



import base64
import io
import zlib
from pathlib import Path
from typing import Dict, List, Optional

import cv2
import numpy as np
from PIL import Image

from supervisely.annotation.annotation import Annotation
from supervisely.annotation.label import Label
from supervisely.annotation.obj_class import ObjClass
from supervisely.geometry.bitmap import Bitmap
from supervisely.geometry.point import Point
from supervisely.geometry.point_location import PointLocation, row_col_list_to_points
from supervisely.geometry.polygon import Polygon
from supervisely.geometry.polyline import Polyline
from supervisely.geometry.rectangle import Rectangle
from supervisely.io.fs import file_exists
from supervisely.io.json import dump_json_file, load_json_file
from supervisely.project.project_meta import ProjectMeta
from supervisely.sly_logger import logger
from supervisely.convert.image.image_helper import validate_image_bounds

labelme_shape_types_to_sly_map = {
    "polygon": Polygon,
    "rectangle": Rectangle,
    "line": Polyline,
    "point": Point,
    "circle": Polygon,
    "mask": Bitmap,
    "linestrip": Polyline,
}


def labelme_polygon_to_sly(coords: List[List[int]]) -> Polygon:
    coords = [[row, col] for (col, row) in coords]
    points = row_col_list_to_points(coords, do_round=True)
    return Polygon(points)


def labelme_rectangle_to_sly(coords: List[List[int]]) -> Rectangle:
    left, top = coords[0]
    right, bottom = coords[1]
    return Rectangle(int(top), int(left), int(bottom), int(right))


def labelme_line_to_sly(coords: List[List[int]]) -> Polyline:
    coords = [[row, col] for (col, row) in coords]
    points = row_col_list_to_points(coords, do_round=True)
    return Polyline(points)


def labelme_point_to_sly(coords: List[List[int]]) -> Point:
    col, row = coords[0]
    return Point(int(row), int(col))


def labelme_to_sly_bitmap(origin: List[List[int]], mask: str) -> Bitmap:
    origin = row_col_list_to_points([[row, col] for (col, row) in origin], do_round=True)
    img = Image.open(io.BytesIO(base64.b64decode(mask)))
    mask = np.array(img).astype(np.bool)
    return Bitmap(mask, origin=origin[0] if len(origin) > 0 else None)


def labelme_circle_to_sly(coords: List[List[int]]) -> Polygon:
    if len(coords) != 2:
        raise ValueError("Circle must have exactly 2 points: center and point on the circle.")
    if len(coords[0]) != 2 or len(coords[1]) != 2:
        raise ValueError("Each point must have exactly 2 coordinates: row and column.")
    center_col, center_row = int(coords[0][0]), int(coords[0][1])
    point_col, point_row = int(coords[1][0]), int(coords[1][1])
    radius = int(np.sqrt((center_row - point_row) ** 2 + (center_col - point_col) ** 2))
    points = []
    for i in range(0, 360, 10):
        col = int(center_col + radius * np.cos(np.radians(i)))
        row = int(center_row + radius * np.sin(np.radians(i)))
        points.append(PointLocation(row, col))
    return Polygon(points)


labelme_shape_types_to_convert_func = {
    "polygon": labelme_polygon_to_sly,
    "rectangle": labelme_rectangle_to_sly,
    "line": labelme_line_to_sly,
    "point": labelme_point_to_sly,
    "circle": labelme_circle_to_sly,
    "mask": labelme_to_sly_bitmap,
    "linestrip": labelme_line_to_sly,
}


def convert_labelme_to_sly(shape: Dict, obj_cls: ObjClass) -> Optional[Label]:
    try:
        shape_type = shape.get("shape_type")
        if shape_type is None:
            raise ValueError("Shape type is not specified.")
        convert_func = labelme_shape_types_to_convert_func.get(shape_type)
        if convert_func is None:
            raise ValueError(f"Unsupported shape type: {shape_type}")
        coords = shape.get("points")
        if coords is None:
            raise ValueError("Shape coordinates are not specified. Skipping.")
        if shape_type == "mask":
            mask = shape.get("mask")
            geometry = convert_func(origin=coords, mask=mask)
        else:
            geometry = convert_func(coords)
        return Label(geometry, obj_cls)
    except Exception as e:
        logger.warn(f"Failed to convert shape: {shape_type}. Reason: {repr(e)}")
        return None


def decode_and_save_image_data(image_data: str, image_path: str) -> str:
    try:
        imencoded = zlib.decompress(base64.b64decode(image_data))
        n = np.frombuffer(imencoded, np.uint8)
        imdecoded = cv2.imdecode(n, cv2.IMREAD_UNCHANGED)
    except zlib.error:
        # If the string is not compressed, we'll not use zlib.
        img = Image.open(io.BytesIO(base64.b64decode(image_data)))
        imdecoded = np.array(img)

    imdecoded = cv2.cvtColor(imdecoded, cv2.COLOR_RGB2BGR)
    cv2.imwrite(image_path, imdecoded)
    return image_path


def get_image_from_data(ann_path: str, possible_image_path: Optional[str]) -> str:
    if possible_image_path is not None and file_exists(possible_image_path):
        return possible_image_path
    ann_json = load_json_file(ann_path)
    image_path = ann_json.get("imagePath")
    if image_path is None:
        image_path = Path(ann_path).with_suffix(".jpg").as_posix()
    if file_exists(image_path):
        return image_path

    # if contains encoded image data in base64
    image_data = ann_json.get("imageData")
    if image_data is not None:
        return decode_and_save_image_data(image_data, image_path)
    return


def generate_new_cls_name(
    meta: ProjectMeta,
    cls_name: str,
    expected_geometry_type: type,
    i: Optional[int] = None,
) -> str:
    new_cls_name = f"{cls_name}_{i}" if i else cls_name
    obj_cls: ObjClass = meta.get_obj_class(new_cls_name)
    if obj_cls is None:
        return new_cls_name
    elif obj_cls.geometry_type == expected_geometry_type:
        return new_cls_name
    return generate_new_cls_name(meta, cls_name, expected_geometry_type, i + 1 if i else 1)


def update_meta_from_labelme_annotation(meta: ProjectMeta, ann_path: str) -> ProjectMeta:
    """Update project meta from LabelMe annotation."""

    raw_json = load_json_file(ann_path)
    shapes = raw_json.get("shapes")
    json_updated = False
    for shape in shapes:
        cls_name = shape.get("label")
        shape_type = shape.get("shape_type")
        geometry_type = labelme_shape_types_to_sly_map.get(shape_type)
        if geometry_type is None:
            logger.warn(f"Unsupported shape type: {shape_type}. Please, contact support.")
        if not cls_name or not shape_type:
            continue
        obj_cls = meta.get_obj_class(cls_name)
        if obj_cls is None:
            obj_cls = ObjClass(cls_name, geometry_type)
            meta = meta.add_obj_class(obj_cls)
        elif obj_cls.geometry_type != geometry_type:
            new_cls_name = generate_new_cls_name(meta, cls_name, geometry_type)
            if new_cls_name != cls_name:
                logger.warn(
                    f"{ann_path}: shape type mismatch for class '{cls_name}'. Renamed to '{new_cls_name}'"
                )
                shape["label"] = new_cls_name
                json_updated = True
            if not meta.obj_classes.has_key(new_cls_name):
                obj_cls = ObjClass(new_cls_name, geometry_type)
                meta = meta.add_obj_class(obj_cls)
    if json_updated:
        dump_json_file(raw_json, ann_path)

    return meta


def create_supervisely_annotation(
    item,
    project_meta: ProjectMeta,
    renamed_classes: Dict[str, str],
) -> Annotation:
    ann = Annotation.from_img_path(item.path)
    h, w = ann.img_size
    if item.ann_data is None:
        return ann
    raw_json = load_json_file(item.ann_data)
    if raw_json.get("imageHeight") != h or raw_json.get("imageWidth") != w:
        logger.warn("Image size in annotation does not match the actual image size. Skipping.")
        return ann

    shapes = raw_json.get("shapes")
    labels = []
    for shape in shapes:
        cls_name = shape.get("label")
        cls_name = renamed_classes.get(cls_name, cls_name)
        obj_class = project_meta.get_obj_class(cls_name)
        if obj_class is None:
            logger.warn(f"Object class '{cls_name}' not found in project meta. Skipping.")
            continue

        label = convert_labelme_to_sly(shape, obj_class)
        if label is not None:
            labels.append(label)
    labels = validate_image_bounds(labels, Rectangle.from_size(ann.img_size))
    ann = ann.add_labels(labels)

    return ann



import uuid
from collections import OrderedDict
from logging import Logger
from typing import Callable, Dict, Generator, List, Optional
from typing import OrderedDict as OrderedDictType

import numpy as np

import supervisely as sly
from supervisely.api.module_api import ApiField
from supervisely.geometry.geometry import Geometry
from supervisely.video_annotation.video_figure import VideoFigure


class TrackerInterface:
    def __init__(
        self,
        context,
        api,
        load_all_frames=False,
        notify_in_predict=False,
        per_point_polygon_tracking=True,
        frame_loader: Callable[[sly.Api, int, int], np.ndarray] = None,
        frames_loader: Callable[[sly.Api, int, List[int]], List[np.ndarray]] = None,
        should_notify: bool = True,
    ):
        self.api: sly.Api = api
        self.logger: Logger = api.logger
        self.frame_index = context["frameIndex"]
        self.frames_count = context["frames"]

        self.track_id = context["trackId"]
        self.video_id = context["videoId"]
        self.object_ids = list(context["objectIds"])
        self.figure_ids = list(context["figureIds"])
        self.direction = context["direction"]
        self.should_notify = should_notify

        # all geometries
        self.stop = len(self.figure_ids) * self.frames_count
        self.global_pos = 0
        self.global_stop_indicatior = False

        self.geometries: OrderedDictType[int, Geometry] = OrderedDict()
        self.frames_indexes: List[int] = []
        self._cur_frames_indexes: List[int] = []
        self._frames: Optional[np.ndarray] = None
        self.load_all_frames = load_all_frames
        self.per_point_polygon_tracking = per_point_polygon_tracking

        # increase self.stop by num of frames will be loaded
        self._add_frames_indexes()

        # increase self.stop by num of points
        self._add_geometries()

        self._hot_cache: Dict[int, np.ndarray] = {}
        self._local_cache_loader = frame_loader
        self._local_cache_frames_loader = frames_loader

        if self.load_all_frames:
            if notify_in_predict:
                self.stop += self.frames_count + 1
            self._load_frames_to_hot_cache()
            self._load_frames()

    def add_object_geometries(self, geometries: List[Geometry], object_id: int, start_fig: int):
        for frame, geometry in zip(self._cur_frames_indexes[1:], geometries):
            if self.global_stop_indicatior:
                self._notify(True, task="stop tracking")
                break
            self.add_object_geometry_on_frame(geometry, object_id, frame)

        self.geometries[start_fig] = geometries[-1]

    def frames_loader_generator(self, batch_size=4) -> Generator[None, None, None]:
        if self.load_all_frames:
            self._cur_frames_indexes = self.frames_indexes
            yield
            return

        self._load_frames_to_hot_cache(
            self.frames_indexes[: min(batch_size + 1, len(self.frames_indexes))]
        )
        ind = self.frames_indexes[0]
        frame = self._load_frame(ind)
        for next_ind_pos, next_ind in enumerate(self.frames_indexes[1:]):
            if next_ind not in self._hot_cache:
                self._load_frames_to_hot_cache(
                    self.frames_indexes[
                        next_ind_pos
                        + 1 : min(next_ind_pos + 1 + batch_size, len(self.frames_indexes))
                    ]
                )
            next_frame = self._load_frame(next_ind)
            self._frames = np.array([frame, next_frame])
            self.frames_count = 1
            self._cur_frames_indexes = [ind, next_ind]
            yield
            frame = next_frame
            ind = next_ind

            if self.global_stop_indicatior:
                self.clear_cache()
                return

    def add_object_geometries_on_frames(
        self,
        geometries: List[Geometry],
        object_ids: List[int],
        frame_indexes: List[int],
        notify: bool = True,
    ):
        def _split(geometries: List[Geometry], object_ids: List[int], frame_indexes: List[int]):
            result = {}
            for geometry, object_id, frame_index in zip(geometries, object_ids, frame_indexes):
                result.setdefault(object_id, []).append((geometry, frame_index))
            return result

        geometries_by_object = _split(geometries, object_ids, frame_indexes)

        for object_id, geometries_frame_indexes in geometries_by_object.items():
            figures_json = [
                {
                    ApiField.OBJECT_ID: object_id,
                    ApiField.GEOMETRY_TYPE: geometry.geometry_name(),
                    ApiField.GEOMETRY: geometry.to_json(),
                    ApiField.META: {ApiField.FRAME: frame_index},
                    ApiField.TRACK_ID: self.track_id,
                }
                for geometry, frame_index in geometries_frame_indexes
            ]
            figures_keys = [uuid.uuid4() for _ in figures_json]
            key_id_map = sly.KeyIdMap()
            self.api.video.figure._append_bulk(
                entity_id=self.video_id,
                figures_json=figures_json,
                figures_keys=figures_keys,
                key_id_map=key_id_map,
            )
            self.logger.debug(f"Added {len(figures_json)} geometries to object #{object_id}")
            if notify:
                self._notify(task="add geometry on frame", pos_increment=len(figures_json))

    def add_object_geometry_on_frame(
        self,
        geometry: Geometry,
        object_id: int,
        frame_ind: int,
        notify: bool = True,
    ):
        self.api.video.figure.create(
            self.video_id,
            object_id,
            frame_ind,
            geometry.to_json(),
            geometry.geometry_name(),
            self.track_id,
        )
        self.logger.debug(f"Added {geometry.geometry_name()} to frame #{frame_ind}")
        if notify:
            self._notify(task="add geometry on frame")

    def clear_cache(self):
        self._hot_cache.clear()

    def _add_geometries(self):
        self.logger.info("Adding geometries.")
        points = 0
        for figure_id in self.figure_ids:
            figure = self.api.video.figure.get_info_by_id(figure_id)
            geometry = sly.deserialize_geometry(figure.geometry_type, figure.geometry)
            self.geometries[figure_id] = geometry

            self.api.logger.debug(f"Added {figure.geometry_type} #{figure_id}")

            # per point track notification
            if isinstance(geometry, sly.Point):
                points += 1
            elif isinstance(geometry, sly.Polygon):
                points += len(geometry.exterior) + len(geometry.interior)
            elif isinstance(geometry, sly.GraphNodes):
                points += len(geometry.nodes.items())
            elif isinstance(geometry, sly.Polyline):
                points += len(geometry.exterior)

        if self.per_point_polygon_tracking:
            if not self.load_all_frames:
                self.stop += points * self.frames_count
            else:
                self.stop += points

        self.logger.info("Geometries added.")
        # TODO: other geometries

    def _add_frames_indexes(self):
        total_frames = self.api.video.get_info_by_id(self.video_id).frames_count
        cur_index = self.frame_index

        while 0 <= cur_index < total_frames and len(self.frames_indexes) < self.frames_count + 1:
            self.frames_indexes.append(cur_index)
            cur_index += 1 if self.direction == "forward" else -1

        if self.load_all_frames:
            self.stop += len(self.frames_indexes)

    def _load_frames_to_hot_cache(self, frames_indexes: List[int] = None):
        if self._local_cache_frames_loader is not None:
            if frames_indexes is None:
                frames_indexes = self.frames_indexes
            frames_to_load = []
            for frame_index in frames_indexes:
                if frame_index not in self._hot_cache:
                    frames_to_load.append(frame_index)
            if len(frames_to_load) == 0:
                return
            self.logger.info(f"Loading {frames_to_load} frames to hot cache.")
            loaded_rgbs = self._local_cache_frames_loader(self.api, self.video_id, frames_to_load)
            for rgb, loaded_frame_index in zip(loaded_rgbs, frames_to_load):
                self._hot_cache[loaded_frame_index] = rgb
            self.logger.info(f"{frames_to_load} frames loaded to hot cache.")

    def _load_frame(self, frame_index):
        if frame_index in self._hot_cache:
            return self._hot_cache[frame_index]
        if self._local_cache_loader is None:
            self._hot_cache[frame_index] = self.api.video.frame.download_np(
                self.video_id, frame_index
            )
            return self._hot_cache[frame_index]
        else:
            return self._local_cache_loader(self.api, self.video_id, frame_index)

    def _load_frames(self):
        rgbs = []
        self.logger.info(f"Loading {len(self.frames_indexes)} frames.")

        for frame_index in self.frames_indexes:
            img_rgb = self._load_frame(frame_index)
            rgbs.append(img_rgb)
            self._notify(task="load frame")
        self._frames = rgbs
        self.logger.info("Frames loaded.")

    def _notify(
        self,
        stop: bool = False,
        fstart: Optional[int] = None,
        fend: Optional[int] = None,
        task: str = "not defined",
        pos_increment: int = 1,
    ):
        if not self.should_notify:
            return

        self.global_pos += pos_increment

        if stop:
            pos = self.stop
        else:
            pos = self.global_pos

        fstart = min(self.frames_indexes) if fstart is None else fstart
        fend = max(self.frames_indexes) if fend is None else fend

        self.logger.debug(f"Task: {task}")
        self.logger.debug(f"Notification status: {pos}/{self.stop}")

        self.global_stop_indicatior = self.api.video.notify_progress(
            self.track_id,
            self.video_id,
            fstart,
            fend,
            pos,
            self.stop,
        )

        self.logger.debug(f"Notification status: stop={self.global_stop_indicatior}")

        if self.global_stop_indicatior and self.global_pos < self.stop:
            self.logger.info("Task stoped by user.")

    @property
    def frames(self) -> np.ndarray:
        return self._frames

    @property
    def frames_with_notification(self) -> np.ndarray:
        """Use this in prediction."""
        self._notify(task="get frames")
        return self._frames



from typing import List

from supervisely.nn.artifacts.artifacts import BaseTrainArtifacts


class HRDA(BaseTrainArtifacts):
    # not enough info to implement

    def __init__(self, team_id: int):
        raise NotImplementedError
        # super().__init__(team_id)
        # self._app_name = "Train HRDA"
        # self._framework_folder = "/HRDA"
        # self._weights_folder = None
        # self._task_type = "semantic segmentation"
        # self._weights_ext = ".pth"
        # self._config_file = "config.py"
        # self._available_task_types: List[str] = ["semantic segmentation"]

    def get_task_id(self, artifacts_folder: str) -> str:
        raise NotImplementedError

    def get_project_name(self, artifacts_folder: str) -> str:
        raise NotImplementedError

    def get_task_type(self, artifacts_folder: str) -> str:
        raise NotImplementedError

    def get_weights_path(self, artifacts_folder: str) -> str:
        raise NotImplementedError

    def get_config_path(self, artifacts_folder: str) -> str:
        raise NotImplementedError



from supervisely.nn.benchmark.semantic_segmentation.benchmark import (
    SemanticSegmentationBenchmark,
)
from supervisely.nn.benchmark.semantic_segmentation.evaluator import (
    SemanticSegmentationEvaluator,
)
from supervisely.nn.benchmark.semantic_segmentation.visualizer import (
    SemanticSegmentationVisualizer,
)



from typing import Any, Dict

from supervisely import Api
from supervisely.app.widgets import (
    Card,
    Container,
    Empty,
    Field,
    Flexbox,
    FolderThumbnail,
    ReportThumbnail,
    Text,
)

PYTORCH_ICON = "https://img.icons8.com/?size=100&id=jH4BpkMnRrU5&format=png&color=000000"
ONNX_ICON = "https://artwork.lfaidata.foundation/projects/onnx/icon/color/onnx-icon-color.png"
TRT_ICON = "https://img.icons8.com/?size=100&id=yqf95864UzeQ&format=png&color=000000"


class TrainingArtifacts:
    title = "Training Artifacts"
    description = "All outputs of the training process will appear here"
    lock_message = "Artifacts will be available after training is completed"

    def __init__(self, app_options: Dict[str, Any]):
        self.display_widgets = []
        self.success_message_text = (
            "Training completed. Training artifacts were uploaded to Team Files. "
            "You can find and open tensorboard logs in the artifacts folder via the "
            "<a href='https://ecosystem.supervisely.com/apps/tensorboard-logs-viewer' target='_blank'>Tensorboard</a> app."
        )
        self.app_options = app_options

        # GUI Components
        self.validator_text = Text("")
        self.validator_text.hide()
        self.display_widgets.extend([self.validator_text])

        # Outputs
        self.artifacts_thumbnail = FolderThumbnail()
        self.artifacts_thumbnail.hide()

        self.artifacts_field = Field(
            title="Artifacts",
            description="Contains all outputs of the training process",
            content=self.artifacts_thumbnail,
        )
        self.artifacts_field.hide()
        self.display_widgets.extend([self.artifacts_field])

        # Optional Model Benchmark
        if app_options.get("model_benchmark", False):
            self.model_benchmark_report_thumbnail = ReportThumbnail()
            self.model_benchmark_report_thumbnail.hide()

            self.model_benchmark_fail_text = Text(
                text="Model evaluation did not finish successfully. Please check the app logs for details.",
                status="error",
            )
            self.model_benchmark_fail_text.hide()

            self.model_benchmark_widgets = Container(
                [self.model_benchmark_report_thumbnail, self.model_benchmark_fail_text]
            )

            self.model_benchmark_report_field = Field(
                title="Model Benchmark",
                description="Evaluation report of the trained model",
                content=self.model_benchmark_widgets,
            )
            self.model_benchmark_report_field.hide()
            self.display_widgets.extend([self.model_benchmark_report_field])
        # -------------------------------- #

        # PyTorch, ONNX, TensorRT demo
        self.inference_demo_field = []
        model_demo = self.app_options.get("demo", None)
        if model_demo is not None:
            pytorch_demo_link = model_demo.get("pytorch", None)
            if pytorch_demo_link is not None:
                pytorch_icon = Field.Icon(image_url=PYTORCH_ICON, bg_color_rgb=[255, 255, 255])
                self.pytorch_instruction = Field(
                    title="PyTorch",
                    description="Open file",
                    description_url=pytorch_demo_link,
                    icon=pytorch_icon,
                    content=Empty(),
                )
                self.pytorch_instruction.hide()
                self.inference_demo_field.extend([self.pytorch_instruction])

            onnx_demo_link = model_demo.get("onnx", None)
            if onnx_demo_link is not None:
                if self.app_options.get("export_onnx_supported", False):
                    onnx_icon = Field.Icon(image_url=ONNX_ICON, bg_color_rgb=[255, 255, 255])
                    self.onnx_instruction = Field(
                        title="ONNX",
                        description="Open file",
                        description_url=onnx_demo_link,
                        icon=onnx_icon,
                        content=Empty(),
                    )
                    self.onnx_instruction.hide()
                    self.inference_demo_field.extend([self.onnx_instruction])

            trt_demo_link = model_demo.get("tensorrt", None)
            if trt_demo_link is not None:
                if self.app_options.get("export_tensorrt_supported", False):
                    trt_icon = Field.Icon(image_url=TRT_ICON, bg_color_rgb=[255, 255, 255])
                    self.trt_instruction = Field(
                        title="TensorRT",
                        description="Open file",
                        description_url=trt_demo_link,
                        icon=trt_icon,
                        content=Empty(),
                    )
                    self.trt_instruction.hide()
                    self.inference_demo_field.extend([self.trt_instruction])

            demo_overview_link = model_demo.get("overview", None)
            self.inference_demo_field = Field(
                title="How to run inference",
                description="Instructions on how to use your checkpoints outside of Supervisely Platform",
                content=Flexbox(self.inference_demo_field),
                title_url=demo_overview_link,
            )
            self.inference_demo_field.hide()
            self.display_widgets.extend([self.inference_demo_field])
        # -------------------------------- #

        self.container = Container(self.display_widgets)
        self.card = Card(
            title=self.title,
            description=self.description,
            content=self.container,
            lock_message=self.lock_message,
        )
        self.card.lock()

    @property
    def widgets_to_disable(self) -> list:
        return []

    def validate_step(self) -> bool:
        return True



# coding: utf-8

import asyncio
import os
import re
from typing import Callable, Dict, List, Literal, Optional, Tuple, Union

import numpy as np
from tqdm import tqdm

from supervisely._utils import batched
from supervisely.annotation.obj_class import ObjClass
from supervisely.api.entity_annotation.entity_annotation_api import EntityAnnotationAPI
from supervisely.api.module_api import ApiField
from supervisely.geometry.any_geometry import AnyGeometry
from supervisely.geometry.mask_3d import Mask3D
from supervisely.io.fs import (
    change_directory_at_index,
    dir_exists,
    file_exists,
    get_file_name,
    list_files,
    silent_remove,
)
from supervisely.io.json import load_json_file
from supervisely.project.project_meta import ProjectMeta
from supervisely.sly_logger import logger
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.volume import stl_converter
from supervisely.volume_annotation.volume_annotation import VolumeAnnotation
from supervisely.volume_annotation.volume_figure import VolumeFigure
from supervisely.volume_annotation.volume_object import VolumeObject
from supervisely.volume_annotation.volume_object_collection import (
    VolumeObjectCollection,
)


class VolumeAnnotationAPI(EntityAnnotationAPI):
    """
    :class:`VolumeAnnotation<supervisely.volume_annotation.volume_annotation.VolumeAnnotation>` for a single volume. :class:`VolumeAnnotationAPI<VolumeAnnotationAPI>` object is immutable.

    :param api: API connection to the server.
    :type api: Api
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        import os
        from dotenv import load_dotenv

        import supervisely as sly

        # Load secrets and create API object from .env file (recommended)
        # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
        if sly.is_development():
            load_dotenv(os.path.expanduser("~/supervisely.env"))
        api = sly.Api.from_env()

        # Pass values into the API constructor (optional, not recommended)
        # api = sly.Api(server_address="https://app.supervise.ly", token="4r47N...xaTatb")

        volume_id = 19581134
        ann_info = api.volume.annotation.download(volume_id)
    """

    _method_download_bulk = "volumes.annotations.bulk.info"
    _entity_ids_str = ApiField.VOLUME_IDS

    def download(self, volume_id: int):
        """
        Download information about VolumeAnnotation by volume ID from API.
        :param volume_id: Volume ID in Supervisely.
        :type volume_id: int
        :return: Information about VolumeAnnotation in json format
        :rtype: :class:`dict`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            from pprint import pprint

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            volume_id = 19581134
            ann_info = api.volume.annotation.download(volume_id)
            print(ann_info)
            # Output:
            # {
            #     'createdAt': '2023-03-29T12:30:37.078Z',
            #     'datasetId': 61803,
            #     'description': '',
            #     'objects': [],
            #     'planes': [],
            #     'spatialFigures': [],
            #     'tags': [{'createdAt': '2023-04-03T13:21:53.368Z',
            #             'id': 12259702,
            #             'labelerLogin': 'almaz',
            #             'name': 'info',
            #             'tagId': 385328,
            #             'updatedAt': '2023-04-03T13:21:53.368Z',
            #             'value': 'age 31'}],
            #     'updatedAt': '2023-03-29T12:30:37.078Z',
            #     'volumeId': 19581134,
            #     'volumeMeta': {
            #             'ACS': 'RAS',
            #             'IJK2WorldMatrix': [0.7617, 0, 0,
            #                                 -194.2384, 0, 0.76171,
            #                                 0, -217.5384, 0,
            #                                 0, 2.5, -347.75,
            #                                 0, 0, 0, 1],
            #             'channelsCount': 1,
            #             'dimensionsIJK': {'x': 512, 'y': 512, 'z': 139},
            #             'intensity': {'max': 3071, 'min': -3024},
            #             'rescaleIntercept': 0,
            #             'rescaleSlope': 1,
            #             'windowCenter': 23.5,
            #             'windowWidth': 6095
            # },
            #     'volumeName': 'CTChest.nrrd'
            # }
        """

        volume_info = self._api.volume.get_info_by_id(volume_id)
        return self._download(volume_info.dataset_id, volume_id)

    def append(
        self, volume_id: int, ann: VolumeAnnotation, key_id_map: KeyIdMap = None, volume_info=None
    ):
        """
        Loads VolumeAnnotation to a given volume ID in the API.

        :param volume_id: Volume ID in Supervisely.
        :type volume_id: int
        :param ann: VolumeAnnotation object.
        :type ann: VolumeAnnotation
        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap, optional
        :return: None
        :rtype: :class:`NoneType`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            volume_id = 19581134
            api.volume.annotation.append(volume_id, volume_ann)
        """
        if ann.spatial_figures:
            figures = ann.figures + ann.spatial_figures
        else:
            figures = ann.figures

        if volume_info is None:
            volume_info = self._api.volume.get_info_by_id(volume_id)
        self._append(
            self._api.volume.tag,
            self._api.volume.object,
            self._api.volume.figure,
            volume_info.project_id,
            volume_info.dataset_id,
            volume_id,
            ann.tags,
            ann.objects,
            figures,
            key_id_map,
        )

    def upload_paths(
        self,
        volume_ids: List[int],
        ann_paths: List[str],
        project_meta: ProjectMeta,
        interpolation_dirs: Optional[List[str]] = None,
        progress_cb: Optional[Union[tqdm, Callable]] = None,
        mask_dirs: Optional[List[str]] = None,
    ) -> None:
        """
        Loads VolumeAnnotations from a given paths to a given volumes IDs in the API. Volumes IDs must be from one dataset.

        :param volume_ids: Volumes IDs in Supervisely.
        :type volume_ids: List[int]
        :param ann_paths: Paths to annotation files
        :type ann_paths: List[str]
        :param project_meta: Input :class:`ProjectMeta<supervisely.project.project_meta.ProjectMeta>` for VolumeAnnotations
        :type project_meta: ProjectMeta
        :param interpolation_dirs: Paths to dirs with interpolation STL files
        :type interpolation_dirs: List[str], optional
        :param progress_cb: Function for tracking download progress
        :type progress_cb: tqdm or callable, optional
        :param mask_dirs: Paths to dirs with Mask3D geometries
        :type mask_dirs: List[str], optional
        :return: None
        :rtype: :class:`NoneType`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            volume_ids = [121236918, 121236919]
            ann_pathes = ['/home/admin/work/supervisely/example/ann1.json', '/home/admin/work/supervisely/example/ann2.json']
            api.volume.annotation.upload_paths(volume_ids, ann_pathes, meta)
        """

        # use in updating project metadata
        project_id = self._api.volume.get_info_by_id(volume_ids[0]).project_id

        if interpolation_dirs is None:
            interpolation_dirs = [None] * len(ann_paths)

        if mask_dirs is None:
            mask_dirs = [None] * len(ann_paths)

        key_id_map = KeyIdMap()
        for volume_id, ann_path, interpolation_dir, mask_dir in zip(
            volume_ids, ann_paths, interpolation_dirs, mask_dirs
        ):
            ann_json = load_json_file(ann_path)
            ann = VolumeAnnotation.from_json(ann_json, project_meta)

            geometries_dict = {}
            stl_paths = []
            nrrd_paths = []
            keep_nrrd_paths = []

            if interpolation_dir is not None and dir_exists(interpolation_dir):
                (
                    stl_paths,
                    nrrd_paths,
                    keep_nrrd_paths,
                ) = self._prepare_convertation_paths(interpolation_dir, ann)

                if len(stl_paths) != 0:
                    stl_converter.to_nrrd(stl_paths, nrrd_paths)
                    ann, project_meta = self._update_on_transfer(
                        "upload", ann, project_meta, nrrd_paths, project_id
                    )

            # list all Mask3D geometries
            if mask_dir is not None and dir_exists(mask_dir):
                mask_paths = list_files(mask_dir, valid_extensions=[".nrrd"])
                geometries_dict.update(Mask3D._bytes_from_nrrd_batch(mask_paths))
                # it is not recommended to change the original composition of the project directory files
                # for this purpose delete the files created during conversion
                for nrrd_path in nrrd_paths:
                    if nrrd_path not in keep_nrrd_paths:
                        silent_remove(nrrd_path)

            # add geometries into spatial figure objects
            for sf in ann.spatial_figures:
                try:
                    geometry_bytes = geometries_dict[sf.key().hex]
                    mask3d = Mask3D.from_bytes(geometry_bytes)
                    sf._set_3d_geometry(mask3d)
                except (KeyError, TypeError) as e:
                    if isinstance(e, TypeError):
                        logger.warning(
                            f"Skipping spatial figure for class '{sf.volume_object.obj_class.name}': {str(e)}"
                        )
                    # skip figures that doesn't need to update geometry
                    # for example for old geometries that are stored in JSON (KeyError)
                    continue

            self.append(volume_id, ann, key_id_map)

            if progress_cb is not None:
                progress_cb(1)

    def _update_on_transfer(
        self,
        transfer_type: Literal["download", "upload"],
        ann: VolumeAnnotation,
        project_meta: ProjectMeta,
        nrrd_paths: List[str],
        project_id: Optional[int] = None,
    ) -> Tuple[VolumeAnnotation, ProjectMeta]:
        """
        Create new ObjClass and VolumeFigure annotations for converted STL.
        Replace ClosedMeshSurface spatial figures with Mask3D.
        Update the ann, project_meta, and key_id_map.

        :param transfer_type: Defines the process during which the update will be performed ("download" or "upload").
        :type transfer_type: Literal["download", "upload"]
        :param ann: The VolumeAnnotation object to update.
        :type ann: VolumeAnnotation
        :param project_meta: The ProjectMeta object.
        :type project_meta: ProjectMeta
        :param nrrd_paths: Paths to the converted NRRD files from STL.
        :type nrrd_paths: List[str]
        :param project_id: The Project ID to update metadata on upload (optional).
        :type project_id: int, optional
        :return: A tuple containing the updated ann and project_meta objects.
        :rtype: Tuple[VolumeAnnotation, ProjectMeta]
        """

        for nrrd_path in nrrd_paths:
            object_key = None

            # searching connection between interpolation and spatial figure in annotations and set its object_key
            for sf in ann.spatial_figures:
                if sf.key().hex == get_file_name(nrrd_path):
                    object_key = sf.parent_object.key()
                    break

            if object_key:
                for obj in ann.objects:
                    if obj.key() == object_key:
                        class_title = obj.obj_class.name
                        break

            else:
                raise Exception(
                    f"Can't find volume object for Mask3D from unknown file '{nrrd_path}'. Please check the project structure."
                )

            class_created = False
            new_obj_class = project_meta.get_obj_class(f"{class_title}_mask_3d")

            if new_obj_class is None:
                new_obj_class = ObjClass(f"{class_title}_mask_3d", Mask3D)
                project_meta = project_meta.add_obj_class(new_obj_class)
                class_created = True

            geometry = Mask3D(np.zeros((3, 3, 3), dtype=np.bool_))

            if transfer_type == "download":
                new_object = VolumeObject(new_obj_class, mask_3d=geometry)
            elif transfer_type == "upload":
                if class_created:
                    self._api.project.update_meta(project_id, project_meta)
                new_object = VolumeObject(new_obj_class)
                new_object.figure = VolumeFigure(new_object, geometry, key=sf.key())

            # add new Volume object to VolumeAnnotation with spatial figure
            ann = ann.add_objects([new_object])

            if transfer_type == "download":
                # as a sf.key() changes, we need to rename both files
                os.rename(
                    nrrd_path, f"{os.path.dirname(nrrd_path)}/{new_object.figure.key().hex}.nrrd"
                )
                stl_path = re.sub(r"\.[^.]+$", ".stl", nrrd_path)
                stl_path = change_directory_at_index(stl_path, "interpolation", -3)
                os.rename(
                    stl_path, f"{os.path.dirname(stl_path)}/{new_object.figure.key().hex}.stl"
                )
            # remove STL spatial figure from VolumeAnnotation
            if sf:
                ann.spatial_figures.remove(sf)

        return ann, project_meta

    def append_objects(
        self,
        volume_id: int,
        objects: Union[List[VolumeObject], VolumeObjectCollection],
        key_id_map: Optional[KeyIdMap] = None,
    ) -> None:
        """
        Add new VolumeObjects to a volume annotation in Supervisely project.

        :param volume_id: The ID of the volume.
        :type volume_id: int
        :param objects: New volume objects.
        :type objects: List[VolumeObject] or VolumeObjectCollection
        :param key_id_map: The KeyIdMap (optional).
        :type key_id_map: KeyIdMap, optional
        :return: None
        :rtype: NoneType

        :Usage example:

         .. code-block:: python

            import os
            from dotenv import load_dotenv

            import supervisely as sly

            # Load secrets and create API object from .env file (recommended)
            # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
            if sly.is_development():
               load_dotenv(os.path.expanduser("~/supervisely.env"))
            api = sly.Api.from_env()

            volume_id = 151344
            volume_info = api.volume.get_info_by_id(volume_id)
            mask_3d_path = "data/mask/lung.nrrd"
            lung_obj_class = sly.ObjClass("lung", sly.Mask3D)
            lung = sly.VolumeObject(lung_obj_class, mask_3d=mask_3d_path)
            objects = sly.VolumeObjectCollection([lung])
            api.volume.annotation.append_objects(volume_info.id, objects)
        """

        sf_figures = []
        for volume_object in objects:
            if volume_object.obj_class.geometry_type in (Mask3D, AnyGeometry):
                if isinstance(volume_object.figure.geometry, Mask3D):
                    sf_figures.append(volume_object.figure)

        volume_meta = self._api.volume.get_info_by_id(volume_id).meta
        ann = VolumeAnnotation(volume_meta, objects, spatial_figures=sf_figures)
        self.append(volume_id, ann, key_id_map)

    @staticmethod
    def _prepare_convertation_paths(
        interpolation_dir: str, ann: VolumeAnnotation
    ) -> Tuple[List, List, List]:
        """
        Check dir and create paths for NRRD files if STL files need to be converted.

        :param interpolation_dir: Path to dir with interpolation STL files
        :type interpolation_dir: str
        :param ann: VolumeAnnotation object
        :type ann: VolumeAnnotation
        :return: Paths to STL and NRRD files used in the conversion process
        :rtype: Tuple[List, List, List]
        """
        stl_paths_in = list_files(interpolation_dir, valid_extensions=[".stl"])
        nrrd_paths = []
        stl_paths = []
        keep_nrrd_paths = []  # to keep original composition of the project
        for stl_path in stl_paths_in:
            # check if this is really STL file
            with open(stl_path, "rb") as file:
                header = file.read(84)
                if b"solid" in header:
                    stl_paths.append(stl_path)
                else:
                    continue
            nrrd_path = re.sub(r"\.[^.]+$", ".nrrd", stl_path)
            nrrd_path = change_directory_at_index(nrrd_path, "mask", -3)
            nrrd_paths.append(nrrd_path)
            if file_exists(nrrd_path):
                keep_nrrd_paths.append(nrrd_path)
                # to prevent duplication if STL is already converted to NRRD on export
                for sf in ann.spatial_figures:
                    if sf.key().hex == get_file_name(nrrd_path) and isinstance(sf.geometry, Mask3D):
                        stl_paths.remove(stl_path)
                        nrrd_paths.remove(nrrd_path)
                        keep_nrrd_paths.remove(nrrd_path)
        return stl_paths, nrrd_paths, keep_nrrd_paths

    async def download_async(
        self,
        volume_id: int,
        semaphore: Optional[asyncio.Semaphore] = None,
        integer_coords: bool = True,
        progress_cb: Optional[Union[tqdm, Callable]] = None,
    ) -> Dict:
        """
        Download information about VolumeAnnotation by volume ID from API asynchronously.

        :param volume_id: Volume ID in Supervisely.
        :type volume_id: int
        :param semaphore: Semaphore to limit the number of parallel downloads.
        :type semaphore: asyncio.Semaphore, optional
        :param integer_coords: If True, returns coordinates as integers for objects. If False, returns as floats.
        :type integer_coords: bool, optional
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm or callable, optional
        :return: Information about VolumeAnnotation in json format
        :rtype: :class:`dict`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            volume_id = 198702499
            loop = sly.utils.get_or_create_event_loop()
            ann_info = loop.run_until_complete(api.volume.annotation.download_async(volume_id))
        """
        return await self.download_bulk_async(
            volume_ids=[volume_id],
            semaphore=semaphore,
            integer_coords=integer_coords,
            progress_cb=progress_cb,
        )

    async def download_bulk_async(
        self,
        volume_ids: List[int],
        semaphore: Optional[asyncio.Semaphore] = None,
        integer_coords: bool = True,
        progress_cb: Optional[Union[tqdm, Callable]] = None,
    ) -> List[Dict]:
        """
        Download information about VolumeAnnotation in bulk by volume IDs from API asynchronously.

        :param volume_ids: List of Volume IDs in Supervisely. All volumes must be from the same dataset.
        :type volume_ids: int
        :param semaphore: Semaphore to limit the number of parallel downloads.
        :type semaphore: asyncio.Semaphore, optional
        :param integer_coords: If True, returns coordinates as integers for objects. If False, returns as floats.
        :type integer_coords: bool, optional
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm or callable, optional
        :return: Information about VolumeAnnotations in json format
        :rtype: :class:`list`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            volume_ids = [198702499, 198702500, 198702501]
            loop = sly.utils.get_or_create_event_loop()
            ann_infos = loop.run_until_complete(api.volume.annotation.download_bulk_async(volume_ids))
        """
        if semaphore is None:
            semaphore = self._api.get_default_semaphore()

        async def fetch_with_semaphore(batch):
            async with semaphore:
                json_data = {
                    self._entity_ids_str: batch,
                    ApiField.INTEGER_COORDS: integer_coords,
                }
                response = await self._api.post_async(
                    self._method_download_bulk,
                    json=json_data,
                )
                if progress_cb is not None:
                    progress_cb(len(batch))
                return response.json()

        tasks = [fetch_with_semaphore(batch) for batch in batched(volume_ids)]
        responses = await asyncio.gather(*tasks)
        json_response = [item for response in responses for item in response]
        return json_response



import functools
import os
import pathlib
import shutil
import tempfile
import threading
import traceback
from collections import namedtuple
from distutils.dir_util import copy_tree

from fastapi import FastAPI
from starlette.routing import Mount
from starlette.staticfiles import StaticFiles
from starlette.templating import _TemplateResponse

import supervisely as sly

_offline_session_uploader = None
_pending_offline_session = None


def get_static_paths_by_mounted_object(mount) -> list:
    StaticPath = namedtuple("StaticPath", ["local_path", "url_path"])
    static_paths = []

    if os.path.exists("static"):
        static_paths.append(StaticPath(local_path=pathlib.Path("static"), url_path="./static"))

    if hasattr(mount, "routes"):
        for current_route in mount.routes:
            if type(current_route) == Mount and type(current_route.app) == FastAPI:
                all_children_paths = get_static_paths_by_mounted_object(current_route)
                for index, current_path in enumerate(all_children_paths):
                    current_url_path = pathlib.Path(
                        str(current_route.path).lstrip("/"),
                        str(current_path.url_path).lstrip("/"),
                    )
                    all_children_paths[index] = StaticPath(
                        local_path=current_path.local_path, url_path=current_url_path
                    )
                static_paths.extend(all_children_paths)
            elif type(current_route) == Mount and type(current_route.app) == StaticFiles:
                static_paths.append(
                    StaticPath(
                        local_path=pathlib.Path(current_route.app.directory),
                        url_path=pathlib.Path(str(current_route.path).lstrip("/")),
                    )
                )

    return static_paths


def dump_statics_to_dir(static_dir_path: pathlib.Path, static_paths: list):
    for current_path in static_paths:
        current_local_path: pathlib.Path = current_path.local_path
        current_url_path: pathlib.Path = static_dir_path / current_path.url_path

        def _filter_static_files(path: pathlib.Path):
            extensions_to_delete = [".py", ".pyc", ".md", ".sh"]
            for dirpath, _, filenames in os.walk(path.as_posix()):
                if filenames:
                    for file in filenames:
                        if os.path.splitext(os.path.basename(file))[1] in extensions_to_delete:
                            filepath = pathlib.Path(dirpath, file)
                            if pathlib.Path.exists(filepath):
                                pathlib.Path.unlink(filepath)

        if current_local_path.is_dir():
            current_url_path.mkdir(parents=True, exist_ok=True)
            copy_tree(
                current_local_path.as_posix(),
                current_url_path.as_posix(),
                preserve_symlinks=True,
            )

            _filter_static_files(current_url_path)


def dump_html_to_dir(static_dir_path, template):
    pathlib.Path(static_dir_path / template.template.name).write_bytes(template.body)


def get_offline_session_files_path(task_id) -> pathlib.Path:
    return pathlib.Path("/", "offline-sessions", str(task_id), "app-template")


def upload_to_supervisely(static_dir_path):
    api_token = sly.env.spawn_api_token(raise_not_found=False) or sly.env.api_token()
    # spawn_api_token - is a token of user, that spawned application.
    # api_token - is a token of user for which application was spawned.
    # It's important to use spawnApiToken, because the application can be spawned
    # by user with higher permissions, than current user (e.g. annotator).
    # For example, if we'lll use annotator's token, we'll get 403 error, when
    # trying to upload files, because annotator doesn't have corresponding permissions.
    api = sly.Api(sly.env.server_address(), api_token)

    team_id = sly.env.team_id()
    task_id = sly.env.task_id(raise_not_found=False)
    task_id = 0000 if task_id is None else task_id
    remote_dir = get_offline_session_files_path(task_id)

    res_remote_dir: str = api.file.upload_directory(
        team_id=team_id,
        local_dir=static_dir_path.as_posix(),
        remote_dir=remote_dir.as_posix(),
        change_name_if_conflict=False,
        replace_if_conflict=True,
    )

    if os.getenv("TASK_ID") is not None:
        api.task.update_meta(id=task_id, data={"templateRootDirectory": res_remote_dir})

    sly.logger.info(f"App files stored in {res_remote_dir} for offline usage")


def dump_files_to_supervisely(app: FastAPI, template_response):
    try:
        if os.getenv("TASK_ID") is None:
            sly.logger.debug(f"Debug mode: saving app files for offline usage is skipped")
            return

        if (
            os.getenv("_SUPERVISELY_OFFLINE_FILES_UPLOADED", "False") == "True"
            and template_response.context.get("request") is not None
        ):
            return
        os.environ["_SUPERVISELY_OFFLINE_FILES_UPLOADED"] = "True"
        sly.logger.info(f"Saving app files for offline usage")

        app_template_path = pathlib.Path(tempfile.mkdtemp())
        app_static_paths = get_static_paths_by_mounted_object(mount=app)
        dump_statics_to_dir(static_dir_path=app_template_path, static_paths=app_static_paths)
        dump_html_to_dir(static_dir_path=app_template_path, template=template_response)

        upload_to_supervisely(static_dir_path=app_template_path)

        shutil.rmtree(app_template_path.as_posix())

    except Exception as ex:
        traceback.print_exc()
        sly.logger.warning(f"Cannot dump files for offline usage, reason: {ex}")
        os.environ["_SUPERVISELY_OFFLINE_FILES_UPLOADED"] = "False"


def _upload_offline_session():
    global _pending_offline_session
    if _pending_offline_session is not None:
        app, template_response = _pending_offline_session
        _pending_offline_session = None
        sly.logger.info(f"Start dumping app UI for offline mode")
        dump_files_to_supervisely(app, template_response)
    if _pending_offline_session is not None:
        _upload_offline_session()


def available_after_shutdown(app: FastAPI):
    def func_layer_wrapper(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            global _offline_session_uploader
            global _pending_offline_session
            template_response = f(*args, **kwargs)
            if not isinstance(template_response, _TemplateResponse):
                return template_response
            try:
                if sly.utils.is_production():
                    _pending_offline_session = (app, template_response)
                    if (
                        _offline_session_uploader is None
                        or not _offline_session_uploader.is_alive()
                    ):
                        _offline_session_uploader = threading.Thread(
                            target=_upload_offline_session,
                            daemon=False,
                        )
                        _offline_session_uploader.start()
            except Exception as ex:
                traceback.print_exc()
                sly.logger.warning(f"Cannot dump files for offline usage, reason: {ex}")

            finally:
                return template_response

        return wrapper

    return func_layer_wrapper



from typing import Dict

from supervisely.nn.benchmark.semantic_segmentation.base_vis_metric import (
    SemanticSegmVisMetric,
)
from supervisely.nn.benchmark.visualization.widgets import ChartWidget, MarkdownWidget


class ConfusionMatrix(SemanticSegmVisMetric):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.clickable = True
        self._keypair_sep = "-"

    @property
    def md(self) -> MarkdownWidget:
        return MarkdownWidget(
            "confusion_matrix",
            "Confusion Matrix",
            text=self.vis_texts.markdown_confusion_matrix,
        )

    @property
    def chart(self) -> ChartWidget:
        chart = ChartWidget("confusion_matrix", self.get_figure())
        chart.set_click_data(
            self.explore_modal_table.id,
            self.get_click_data(),
            chart_click_extra="'getKey': (payload) => `${payload.points[0].y}${'-'}${payload.points[0].x}`, 'keySeparator': '-',",
        )
        return chart

    def get_figure(self):
        import plotly.graph_objects as go  # pylint: disable=import-error

        fig = go.Figure()

        # # Confusion Matrix figure
        confusion_matrix, class_names = self.eval_result.mp.confusion_matrix

        x = class_names
        y = x[::-1].copy()
        if len(x) >= 20:
            text_anns = [[str(el) for el in row] for row in confusion_matrix]
        else:
            text_anns = [
                [
                    f"Predicted: {pred}<br>Ground Truth: {gt}<br> Probability: {confusion_matrix[ig][ip]}"
                    for ip, pred in enumerate(x)
                ]
                for ig, gt in enumerate(y)
            ]

        fig.add_trace(
            go.Heatmap(
                z=confusion_matrix,
                x=x,
                y=y,
                colorscale="Viridis",
                showscale=False,
                text=text_anns,
                hoverinfo="text",
            )
        )

        fig.update_layout(xaxis_title="Predicted", yaxis_title="Ground Truth")
        if len(x) <= 20:
            fig.update_layout(width=600, height=600)
        return fig

    def get_click_data(self) -> Dict:
        res = dict(projectMeta=self.eval_result.pred_project_meta.to_json())
        res["layoutTemplate"] = [None, None, None]
        res["clickData"] = {}

        _, class_names = self.eval_result.mp.confusion_matrix
        for ig, gt_key in enumerate(class_names):
            for ip, pred_key in enumerate(class_names):
                key = f"{gt_key}{self._keypair_sep}{pred_key}"
                res["clickData"][key] = {}
                res["clickData"][key]["imagesIds"] = []

                cmat_key = str(ig) + "_" + str(ip)
                for name in self.eval_result.mp.cmat_cell_img_names[cmat_key]:
                    gt_img_id = self.eval_result.images_map[name]
                    pred_img_id = self.eval_result.matched_pair_data[gt_img_id].pred_image_info.id
                    res["clickData"][key]["imagesIds"].append(pred_img_id)
                title = f"Confusion Matrix. GT: '{gt_key}' ― Predicted: '{pred_key}'"
                res["clickData"][key]["title"] = title

        return res



from __future__ import absolute_import

import numpy as np

from ..utils.kalman_filter import KalmanFilterXYAH as KalmanFilter
from .deep_sort import iou_matching, linear_assignment


class TrackState:
    """
    Enumeration type for the single target track state. Newly created tracks are
    classified as `tentative` until enough evidence has been collected. Then,
    the track state is changed to `confirmed`. Tracks that are no longer alive
    are classified as `deleted` to mark them for removal from the set of active
    tracks.

    """

    Tentative = 1
    Confirmed = 2
    Deleted = 3


class Track:
    """
    A single target track with state space `(x, y, a, h)` and associated
    velocities, where `(x, y)` is the center of the bounding box, `a` is the
    aspect ratio and `h` is the height.

    Parameters
    ----------
    mean : ndarray
        Mean vector of the initial state distribution.
    covariance : ndarray
        Covariance matrix of the initial state distribution.
    track_id : int
        A unique track identifier.
    n_init : int
        Number of consecutive detections before the track is confirmed. The
        track state is set to `Deleted` if a miss occurs within the first
        `n_init` frames.
    max_age : int
        The maximum number of consecutive misses before the track state is
        set to `Deleted`.
    feature : Optional[ndarray]
        Feature vector of the detection this track originates from. If not None,
        this feature is added to the `features` cache.

    Attributes
    ----------
    mean : ndarray
        Mean vector of the initial state distribution.
    covariance : ndarray
        Covariance matrix of the initial state distribution.
    track_id : int
        A unique track identifier.
    hits : int
        Total number of measurement updates.
    age : int
        Total number of frames since first occurance.
    time_since_update : int
        Total number of frames since last measurement update.
    state : TrackState
        The current track state.
    features : List[ndarray]
        A cache of features. On each measurement update, the associated feature
        vector is added to this list.

    """

    def __init__(self, mean, covariance, track_id, n_init, max_age, feature=None):
        self.mean = mean
        self.covariance = covariance
        self.track_id = track_id
        self.hits = 1
        self.age = 1
        self.time_since_update = 0

        self.state = TrackState.Tentative
        self.features = []
        if feature is not None:
            self.features.append(feature)

        self._n_init = n_init
        self._max_age = max_age

    def to_tlwh(self):
        """Get current position in bounding box format `(top left x, top left y,
        width, height)`.

        Returns
        -------
        ndarray
            The bounding box.

        """
        ret = self.mean[:4].copy()
        ret[2] *= ret[3]
        ret[:2] -= ret[2:] / 2
        return ret

    def to_tlbr(self):
        """Get current position in bounding box format `(min x, miny, max x,
        max y)`.

        Returns
        -------
        ndarray
            The bounding box.

        """
        ret = self.to_tlwh()
        ret[2:] = ret[:2] + ret[2:]
        return ret

    def predict(self, kf):
        """Propagate the state distribution to the current time step using a
        Kalman filter prediction step.

        Parameters
        ----------
        kf : kalman_filter.KalmanFilter
            The Kalman filter.

        """
        self.mean, self.covariance = kf.predict(self.mean, self.covariance)
        self.age += 1
        self.time_since_update += 1

    def update(self, kf, detection):
        """Perform Kalman filter measurement update step and update the feature
        cache.

        Parameters
        ----------
        kf : kalman_filter.KalmanFilter
            The Kalman filter.
        detection : Detection
            The associated detection.

        """
        self.mean, self.covariance = kf.update(self.mean, self.covariance, detection.to_xyah())
        self.features.append(detection.feature)

        self.hits += 1
        self.time_since_update = 0
        if self.state == TrackState.Tentative and self.hits >= self._n_init:
            self.state = TrackState.Confirmed

    def mark_missed(self):
        """Mark this track as missed (no association at the current time step)."""
        if self.state == TrackState.Tentative:
            self.state = TrackState.Deleted
        elif self.time_since_update > self._max_age:
            self.state = TrackState.Deleted

    def is_tentative(self):
        """Returns True if this track is tentative (unconfirmed)."""
        return self.state == TrackState.Tentative

    def is_confirmed(self):
        """Returns True if this track is confirmed."""
        return self.state == TrackState.Confirmed

    def is_deleted(self):
        """Returns True if this track is dead and should be deleted."""
        return self.state == TrackState.Deleted


class Tracker:
    """
    This is the multi-target tracker.

    Parameters
    ----------
    metric : nn_matching.NearestNeighborDistanceMetric
        A distance metric for measurement-to-track association.
    max_age : int
        Maximum number of missed misses before a track is deleted.
    n_init : int
        Number of consecutive detections before the track is confirmed. The
        track state is set to `Deleted` if a miss occurs within the first
        `n_init` frames.

    Attributes
    ----------
    metric : nn_matching.NearestNeighborDistanceMetric
        The distance metric used for measurement to track association.
    max_age : int
        Maximum number of missed misses before a track is deleted.
    n_init : int
        Number of frames that a track remains in initialization phase.
    kf : kalman_filter.KalmanFilter
        A Kalman filter to filter target trajectories in image space.
    tracks : List[Track]
        The list of active tracks at the current time step.

    """

    def __init__(self, metric, max_iou_distance=0.7, max_age=30, n_init=3):
        self.metric = metric
        self.max_iou_distance = max_iou_distance
        self.max_age = max_age
        self.n_init = n_init

        self.kf = KalmanFilter()
        self.tracks = []
        self._next_id = 1

    def predict(self):
        """Propagate track state distributions one time step forward.

        This function should be called once every time step, before `update`.
        """
        for track in self.tracks:
            track.predict(self.kf)

    def update(self, detections):
        """Perform measurement update and track management.

        Parameters
        ----------
        detections : List[deep_sort.detection.Detection]
            A list of detections at the current time step.

        """
        # Run matching cascade.
        matches, unmatched_tracks, unmatched_detections = self._match(detections)

        # Update track set.
        for track_idx, detection_idx in matches:
            self.tracks[track_idx].update(self.kf, detections[detection_idx])
        for track_idx in unmatched_tracks:
            self.tracks[track_idx].mark_missed()
        for detection_idx in unmatched_detections:
            self._initiate_track(detections[detection_idx])
        self.tracks = [t for t in self.tracks if not t.is_deleted()]

        # Update distance metric.
        active_targets = [t.track_id for t in self.tracks if t.is_confirmed()]
        features, targets = [], []
        for track in self.tracks:
            if not track.is_confirmed():
                continue
            features += track.features
            targets += [track.track_id for _ in track.features]
            track.features = []
        self.metric.partial_fit(np.asarray(features), np.asarray(targets), active_targets)

    def _match(self, detections):

        def gated_metric(tracks, dets, track_indices, detection_indices):
            features = np.array([dets[i].feature for i in detection_indices])
            targets = np.array([tracks[i].track_id for i in track_indices])
            cost_matrix = self.metric.distance(features, targets)
            cost_matrix = linear_assignment.gate_cost_matrix(
                self.kf, cost_matrix, tracks, dets, track_indices, detection_indices
            )

            return cost_matrix

        # Split track set into confirmed and unconfirmed tracks.
        confirmed_tracks = [i for i, t in enumerate(self.tracks) if t.is_confirmed()]
        unconfirmed_tracks = [i for i, t in enumerate(self.tracks) if not t.is_confirmed()]

        # Associate confirmed tracks using appearance features.
        matches_a, unmatched_tracks_a, unmatched_detections = linear_assignment.matching_cascade(
            gated_metric,
            self.metric.matching_threshold,
            self.max_age,
            self.tracks,
            detections,
            confirmed_tracks,
        )

        # Associate remaining tracks together with unconfirmed tracks using IOU.
        iou_track_candidates = unconfirmed_tracks + [
            k for k in unmatched_tracks_a if self.tracks[k].time_since_update == 1
        ]
        unmatched_tracks_a = [
            k for k in unmatched_tracks_a if self.tracks[k].time_since_update != 1
        ]
        matches_b, unmatched_tracks_b, unmatched_detections = linear_assignment.min_cost_matching(
            iou_matching.iou_cost,
            self.max_iou_distance,
            self.tracks,
            detections,
            iou_track_candidates,
            unmatched_detections,
        )

        matches = matches_a + matches_b
        unmatched_tracks = list(set(unmatched_tracks_a + unmatched_tracks_b))
        return matches, unmatched_tracks, unmatched_detections

    def _initiate_track(self, detection):
        mean, covariance = self.kf.initiate(detection.to_xyah())
        self.tracks.append(
            Track(mean, covariance, self._next_id, self.n_init, self.max_age, detection.feature)
        )
        self._next_id += 1



from supervisely.nn.training.loggers.base_train_logger import BaseTrainLogger
tensorboard_installed = False
try:
    from tensorboardX import SummaryWriter
    tensorboard_installed = True
except ImportError:
    pass


class TensorboardLogger(BaseTrainLogger):
    def __init__(self, log_dir=None):
        if log_dir is None:
            self.log_dir = None
            self.writer = None
        else:
            self.log_dir = log_dir
            self.writer = SummaryWriter(log_dir)
        super().__init__()

    def set_log_dir(self, log_dir):
        self.log_dir = log_dir
        self.writer = SummaryWriter(log_dir)
    
    def close(self):
        if self.writer is not None:
            self.writer.close()
        self.writer = None

    def log(self, logs: dict, idx: int):
        for key, value in logs.items():
            if isinstance(value, (int, float)):
                self.writer.add_scalar(key, value, idx)
        self.writer.flush()

    def log_step(self, logs: dict):
        self.log(logs, self.step_idx)

    def log_epoch(self, logs: dict):
        self.log(logs, self.epoch_idx)



# coding: utf-8

# docs
from __future__ import annotations

from typing import Any, Dict, List, Optional, Tuple

from supervisely.api.module_api import ApiField
from supervisely.collection.key_indexed_collection import KeyIndexedCollection
from supervisely.video_annotation.frame import Frame
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.video_annotation.video_figure import VideoFigure
from supervisely.video_annotation.video_object_collection import VideoObjectCollection


class FrameCollection(KeyIndexedCollection):
    """
    Collection with :class:`Frame<supervisely.video_annotation.frame.Frame>` instances. :class:`FrameCollection<FrameCollection>` object is immutable.

    :Usage example:

     .. code-block:: python

        import supervisely as sly

        # Create two frames for collection
        fr_index_1 = 7
        frame_1 = sly.Frame(fr_index_1)
        fr_index_2 = 10
        frame_2 = sly.Frame(fr_index_2)

        # Create FrameCollection
        fr_collection = sly.FrameCollection([frame_1, frame_2])
        print(fr_collection.to_json())
        # Output: [
        #     {
        #         "index": 7,
        #         "figures": []
        #     },
        #     {
        #         "index": 10,
        #         "figures": []
        #     }
        # ]

        # Add item to FrameCollection
        frame_3 = sly.Frame(12)
        # Remember that TagCollection is immutable, and we need to assign new instance of TagCollection to a new variable
        new_fr_collection = fr_collection.add(frame_3)
        print(new_fr_collection.to_json())
        # Output: [
        #     {
        #         "index": 7,
        #         "figures": []
        #     },
        #     {
        #         "index": 10,
        #         "figures": []
        #     },
        #     {
        #         "index": 12,
        #         "figures": []
        #     }
        # ]

        # You can also add multiple items to collection
        frame_3 = sly.Frame(12)
        frame_4 = sly.Frame(15)
        # Remember that TagCollection is immutable, and we need to assign new instance of TagCollection to a new variable
        new_fr_collection = fr_collection.add_items([frame_3, frame_4])
        print(new_fr_collection.to_json())
        # Output: [
        #     {
        #         "index": 7,
        #         "figures": []
        #     },
        #     {
        #         "index": 10,
        #         "figures": []
        #     },
        #     {
        #         "index": 12,
        #         "figures": []
        #     },
        #     {
        #         "index": 15,
        #         "figures": []
        #     }
        # ]

        # Has key, checks if given key exist in collection
        fr_collection.has_key(7)
        # Output: True

        # Intersection, finds intersection of given list of instances with collection items
        frame_1 = sly.Frame(7)
        frame_2 = sly.Frame(10)
        fr_collection = sly.FrameCollection([frame_1, frame_2])
        frame_3 = sly.Frame(12)
        frames_intersections = fr_collection.intersection([frame_3])
        print(frames_intersections.to_json())
        # Output: []

        frames_intersections = fr_collection.intersection([frame_2])
        print(frames_intersections.to_json())
        # Output: [
        #     {
        #         "index": 10,
        #         "figures": []
        #     }
        # ]

        # Note, two frames with the same index values are not equal
        frame_4 = sly.Frame(10)
        frames_intersections = fr_collection.intersection([frame_4])
        # Output:
        # ValueError: Different values for the same key 10

        # Difference, finds difference between collection and given list of Frames
        frames_difference = fr_collection.difference([frame_2])
        print(frames_difference.to_json())
        # Output: [
        #     {
        #         "index": 7,
        #         "figures": []
        #     }
        # ]

        # Merge, merges collection and given list of FrameCollection
        frame_3 = sly.Frame(12)
        frame_4 = sly.Frame(15)
        over_collection = sly.FrameCollection([frame_3, frame_4])
        merged_collection = fr_collection.merge(over_collection)
        print(merged_collection.to_json())
        # Output: [
        #     {
        #         "index": 12,
        #         "figures": []
        #     },
        #     {
        #         "index": 15,
        #         "figures": []
        #     },
        #     {
        #         "index": 7,
        #         "figures": []
        #     },
        #     {
        #         "index": 10,
        #         "figures": []
        #     }
        # ]
    """

    item_type = Frame

    def get(self, key: str, default: Optional[Any] = None) -> Frame:
        """
        Get item from collection with given key(name) and set a default if item does not exist.

        :param key: Name of Frame in collection.
        :type key: str
        :param default: The value that is returned if there is no key in the collection.
        :type default:  Optional[Any]
        :return: :class:`Frame<supervisely.video_annotation.frame.Frame>`, :class:`Slice<supervisely.volume_annotation.slice.Slice>` or :class:`PointcloudEpisodeFrame<supervisely.pointcloud_annotation.pointcloud_episode_frame.PointcloudEpisodeFrame>` object
        :rtype: :class:`KeyObject<KeyObject>`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            frame_index = 7
            geometry = sly.Rectangle(0, 0, 100, 100)
            class_car = sly.ObjClass('car', sly.Rectangle)
            object_car = sly.VideoObject(class_car)
            figure_car = sly.VideoFigure(object_car, geometry, frame_index)

            frame = sly.Frame(frame_index, figures=[figure_car])
            frame_collection = sly.FrameCollection([frame])

            item = frame_collection.get(frame_index)
            pprint(item.to_json())
            # Output: {
            #     "figures": [
            #         {
            #         "geometry": {
            #             "points": {
            #             "exterior": [
            #                 [0, 0],
            #                 [100, 100]
            #             ],
            #             "interior": []
            #             }
            #         },
            #         "geometryType": "rectangle",
            #         "key": "713968a7d5384709bc5d4e63cd4535f2",
            #         "objectKey": "3342e68eff3b44dcb75712499265be55"
            #         }
            #     ],
            #     "index": 7
            # }
        """
        return super().get(key, default)

    def to_json(self, key_id_map: KeyIdMap = None) -> List[Dict]:
        """
        Convert the FrameCollection to a list of json dicts. Read more about `Supervisely format <https://docs.supervise.ly/data-organization/00_ann_format_navi>`_.

        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap, optional
        :return: List of dicts in json format
        :rtype: :class:`List[dict]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            frame_1 = sly.Frame(7)
            frame_2 = sly.Frame(10)

            # Create FrameCollection
            fr_collection = sly.FrameCollection([frame_1, frame_2])
            print(fr_collection.to_json())
            # Output: [
            #     {
            #         "index": 7,
            #         "figures": []
            #     },
            #     {
            #         "index": 10,
            #         "figures": []
            #     }
            # ]
        """
        return [frame.to_json(key_id_map) for frame in self]

    @classmethod
    def from_json(
        cls,
        data: List[Dict],
        objects: VideoObjectCollection,
        frames_count: Optional[int] = None,
        key_id_map: Optional[KeyIdMap] = None,
    ) -> FrameCollection:
        """
        Convert a list of json dicts to FrameCollection. Read more about `Supervisely format <https://docs.supervise.ly/data-organization/00_ann_format_navi>`_.

        :param data: List with dicts in json format.
        :type data: List[dict]
        :param objects: VideoObjectCollection object.
        :type objects: VideoObjectCollection
        :param frames_count: Number of frames in video.
        :type frames_count: int, optional
        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap, optional
        :return: FrameCollection object
        :rtype: :class:`FrameCollection`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            fr_collection_json = [
                {
                    "index": 7,
                    "figures": []
                },
                {
                    "index": 10,
                    "figures": []
                }
            ]

            objects = []
            fr_collection = sly.FrameCollection.from_json(fr_collection_json, objects)
        """
        frames = [
            cls.item_type.from_json(frame_json, objects, frames_count, key_id_map)
            for frame_json in data
        ]
        return cls(frames)

    def __str__(self):
        return "Frames:\n" + super(FrameCollection, self).__str__()

    @property
    def figures(self) -> List[VideoFigure]:
        """
        Get figures from all frames in collection.

        :return: List of figures from all frames in collection
        :rtype: :class:`List[VideoFigure]<supervisely.video_annotation.video_figure.VideoFigure>`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            fr_index_1 = 7
            geometry = sly.Rectangle(0, 0, 100, 100)
            obj_class_car = sly.ObjClass('car', sly.Rectangle)
            video_object_car = sly.VideoObject(obj_class_car)
            video_figure_car = sly.VideoFigure(video_object_car, geometry, fr_index_1)
            frame_1 = sly.Frame(fr_index_1, figures=[video_figure_car])

            fr_index_2 = 10
            geometry = sly.Rectangle(0, 0, 500, 600)
            obj_class_bus = sly.ObjClass('bus', sly.Rectangle)
            video_object_bus = sly.VideoObject(obj_class_bus)
            video_figure_bus = sly.VideoFigure(video_object_bus, geometry, fr_index_2)
            frame_2 = sly.Frame(fr_index_2, figures=[video_figure_bus])

            fr_collection = sly.FrameCollection([frame_1, frame_2])
            figures = fr_collection.figures
        """
        figures_array = []
        for frame in self:
            figures_array.extend(frame.figures)
        return figures_array

    def get_figures_and_keys(self, key_id_map: KeyIdMap) -> Tuple[List[Dict], List[str]]:
        """
        Get figures from all frames in collection in json format, keys from all figures in frames in collection.

        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap
        :return: Figures from all frames in collection in json format, keys from all figures in frames in collection
        :rtype: :class:`Tuple[list, list]`

        :Usage example:

         .. code-block:: python

            import supervisely as sly
            from supervisely.video_annotation.key_id_map import KeyIdMap

            key_id_map = KeyIdMap()

            fr_index_1 = 7
            geometry = sly.Rectangle(0, 0, 100, 100)
            obj_class_car = sly.ObjClass('car', sly.Rectangle)
            video_object_car = sly.VideoObject(obj_class_car)
            video_figure_car = sly.VideoFigure(video_object_car, geometry, fr_index_1)
            frame_1 = sly.Frame(fr_index_1, figures=[video_figure_car])

            fr_index_2 = 10
            geometry = sly.Rectangle(0, 0, 500, 600)
            obj_class_bus = sly.ObjClass('bus', sly.Rectangle)
            video_object_bus = sly.VideoObject(obj_class_bus)
            video_figure_bus = sly.VideoFigure(video_object_bus, geometry, fr_index_2)
            frame_2 = sly.Frame(fr_index_2, figures=[video_figure_bus])

            fr_collection = sly.FrameCollection([frame_1, frame_2])
            figures, keys = fr_collection.get_figures_and_keys(key_id_map)
            print(keys) # [UUID('0ac041b2-314e-4f6b-9d38-704b341fb383'), UUID('88aa1cb3-b1e3-480f-8ace-6346c9a9daba')]

            print(figures)
            # Output: [
            #     {
            #         "key": "a8cae05d6b8c4a67b18004130941fdec",
            #         "objectKey": "cc9a9475d360481c9753f8ac3c63f8b7",
            #         "geometryType": "rectangle",
            #         "geometry": {
            #             "points": {
            #                 "exterior": [
            #                     [
            #                         0,
            #                         0
            #                     ],
            #                     [
            #                         100,
            #                         100
            #                     ]
            #                 ],
            #                 "interior": []
            #             }
            #         },
            #         "meta": {
            #             "frame": 7
            #         }
            #     },
            #     {
            #         "key": "6e00287acc4644dfb21d67406534080b",
            #         "objectKey": "cad78d53ffc84e69a28f5f8941be9021",
            #         "geometryType": "rectangle",
            #         "geometry": {
            #             "points": {
            #                 "exterior": [
            #                     [
            #                         0,
            #                         0
            #                     ],
            #                     [
            #                         600,
            #                         500
            #                     ]
            #                 ],
            #                 "interior": []
            #             }
            #         },
            #         "meta": {
            #             "frame": 10
            #         }
            #     }
            # ]
        """
        keys = []
        figures_json = []
        for frame in self:
            for figure in frame.figures:
                keys.append(figure.key())
                figure_json = figure.to_json(key_id_map)
                figure_json[ApiField.META] = figure.get_meta()
                figures_json.append(figure_json)
        return figures_json, keys
